/* cpu.c: Signetics 2650 CPU emulation module (platform-independent)

Known/suspected bugs:
 (*) AmiArcadia only:
     Doesn't emulate properly unless compiled without optimizations.
     (SAS/C 6.58 compiler bug?)
*/

#ifdef __VBCC__
    #define AMIGA
#endif
#ifdef AMIGA
    #include "amiga.h"
    #include <proto/locale.h>  // GetCatalogStr()
    #include <proto/dos.h>     // Write()
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
    #define EXEC_TYPES_H
    #include "resource.h"
#endif

#include <stdio.h>             // zprintf()
#include <stdlib.h>            // for malloc(), calloc(), free(), rand()
#include <string.h>            // strcpy()

#include "aa.h"
#define CATCOMP_NUMBERS
#define CATCOMP_BLOCK
#include "aa_strings.h"

// #define SUGGEST
// whether to suggest game optimizations

// #define VERYQUICK
// Why does VERYQUICK make it slower!?
#define QUICK
// whether to ignore some page wrapping issues which don't affect any known game

#define gosub          pushras(); branch();
#define BRANCHCODE     (opcode & 3)
#define CCFIELD        ((psl & PSL_CC) >> 6)
#ifdef VERYQUICK
    #define ONE_BYTE       iar++;
    #define TWO_BYTES      iar += 2;
    #define THREE_BYTES    iar += 3;
#else
    #define ONE_BYTE       iar = (iar & PAGE) + ((iar + 1) & NONPAGE);
    #define TWO_BYTES      iar = (iar & PAGE) + ((iar + 2) & NONPAGE);
    #define THREE_BYTES    iar = (iar & PAGE) + ((iar + 3) & NONPAGE);
#endif
#define REG_CYCLES     if (supercpu) cycles += 3; else cycles += 6;
#define SIX_CYCLES     cycles += 6;
#define NINE_CYCLES    cycles += 9;
#define TWELVE_CYCLES  cycles += 12;
#define FIFTEEN_CYCLES cycles += 15;
#ifdef QUICK
    #define WRAP(x)    (iar + x)
#else
    #define WRAP(x)    ((iar & PAGE) + ((iar + x) & NONPAGE))
#endif
#define OPERAND        (memory[WRAP(1)])
#define GET_RR         rr = (opcode & 3); if (rr && (psl & PSL_RS)) rr += 3;
#define CALMREG        if (reg[0] <= '3') \
                       { calm[0] = reg[0] - '0' + 'A'; calm[1] = EOS;  } else \
                       { calm[0] = reg[0] - '4' + 'A'; calm[1] = '\''; }

/* Known CPU inaccuracies:
   (a) if we #define QUICK and/or VERYQUICK, there are, of course, page
       wrapping issues.
   (b) every access to memory should go through cpuread(), so that the
       correct mirrorring, and correct reading of unreadable addresses,
       occurs.
   However, this is really more of theoretical than practical interest;
   ie. no known games depend on this behaviour, and implementing this
   behaviour would slow down the CPU emulation considerably.
     The only known case is the bizarre case where we need to resolve an
   indirect address in an unreadable area (eg. LODA,r0 *$17B0 for
   Galaxia). This is probably a suspect dump. */

EXPORT const int table_sz[2][256] = {
{   1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, //   0.. 15 (LOD)
    1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, //  16.. 31 (SPSW/RETC/BCTR/BCTA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, //  32.. 47 (EOR)
    1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, //  48.. 63 (REDC/RETE/BSTR/BSTA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, //  64.. 79 (HALT/AND)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, //  80.. 95 (RRR/REDE/BRNR/BRNA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, //  96..111 (IOR)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, // 112..127 (REDD/CPSW/PPSW/BSNR/BSNA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, // 128..143 (ADD)
    1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, // 144..159 (LPSW/DAR/BCFR/ZBRR/BCFA/BXA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, // 160..175 (SUB)
    1, 1, 1, 1, 2, 2, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, // 176..191 (WRTC/TPSW/BSFR/ZBSR/BSFA/BSXA)
    1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, // 192..207 (NOP/STR)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, // 208..223 (RRL/WRTE/BIRR/BIRA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, // 224..239 (COM)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3  // 240..255 (WRTD/TMI/BDRR/BDRA)
},
{   1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, //   0.. 15 (LOD)
    3, 3, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, //  16.. 31 (LPSL/SPSL/SPSW/RETC/BCTR/BCTA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, //  32.. 47 (EOR)
    1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, //  48.. 63 (REDC/RETE/BSTR/BSTA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, //  64.. 79 (HALT/AND)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, //  80.. 95 (RRR/REDE/BRNR/BRNA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, //  96..111 (IOR)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, // 112..127 (REDD/CPSW/PPSW/BSNR/BSNA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, // 128..143 (ADD)
    1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, // 144..159 (LPSW/DAR/BCFR/ZBRR/BCFA/BXA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, // 160..175 (SUB)
    1, 1, 1, 1, 2, 2, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, // 176..191 (WRTC/TPSW/BSFR/ZBSR/BSFA/BSXA)
    1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, // 192..207 (NOP/STR)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, // 208..223 (RRL/WRTE/BIRR/BIRA)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, // 224..239 (COM)
    1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3  // 240..255 (WRTD/TMI/BDRR/BDRA)
} }, table_clockperiods[2][256] = {
{   6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, //   0..15  (LOD)
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, //  16..31
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, //  32..47  (EOR)
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, //  48..63
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, //  64..79
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, //  80..95  (RRR/REDE/BRNR/BRNA)
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, //  96..111 (IOR)
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, // 112..127
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, // 128..143 (ADD)
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, // 144..159
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, // 160..175 (SUB)
    6, 6, 6, 6, 9, 9, 6, 6, 9, 9, 9, 9,  9,  9,  9,  9, // 176..191
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, // 192..207
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, // 208..223 (RRL/WRTE/BIRR/BIRA)
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, // 224..239 (COM)
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9  // 240..255 (WRTD/TMI/BDRR/BDRA)
},
{   3, 3, 3, 3, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, //   0..15  (LOD)
   12,12, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, //  16..31
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, //  32..47  (EOR)
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, //  48..63
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, //  64..79
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, //  80..95  (RRR/REDE/BRNR/BRNA)
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, //  96..111 (IOR)
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, // 112..127
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, // 128..143 (ADD)
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, // 144..159
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, // 160..175 (SUB)
    6, 6, 6, 6, 9, 9, 6, 6, 9, 9, 9, 9,  9,  9,  9,  9, // 176..191
    6, 3, 3, 3, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, // 192..207
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9, // 208..223 (RRL/WRTE/BIRR/BIRA)
    6, 6, 6, 6, 6, 6, 6, 6, 9, 9, 9, 9, 12, 12, 12, 12, // 224..239 (COM)
    6, 6, 6, 6, 9, 9, 9, 9, 9, 9, 9, 9,  9,  9,  9,  9  // 240..255 (WRTD/TMI/BDRR/BDRA)
} };
// Indirect addressing instructions require 15 or 18 cycles (6 more than usual).

/***************************************************************
 * handy table to build PC relative offsets
 * from HR (holding register)
 ***************************************************************/
MODULE const int S2650_relative[0x100] =
{     0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
     16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
     32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
     48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
    -64,-63,-62,-61,-60,-59,-58,-57,-56,-55,-54,-53,-52,-51,-50,-49,
    -48,-47,-46,-45,-44,-43,-42,-41,-40,-39,-38,-37,-36,-35,-34,-33,
    -32,-31,-30,-29,-28,-27,-26,-25,-24,-23,-22,-21,-20,-19,-18,-17,
    -16,-15,-14,-13,-12,-11,-10, -9, -8, -7, -6, -5, -4, -3, -2, -1,
      0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11, 12, 13, 14, 15,
     16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31,
     32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,
     48, 49, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
    -64,-63,-62,-61,-60,-59,-58,-57,-56,-55,-54,-53,-52,-51,-50,-49,
    -48,-47,-46,-45,-44,-43,-42,-41,-40,-39,-38,-37,-36,-35,-34,-33,
    -32,-31,-30,-29,-28,-27,-26,-25,-24,-23,-22,-21,-20,-19,-18,-17,
    -16,-15,-14,-13,-12,-11,-10, -9, -8, -7, -6, -5, -4, -3, -2, -1
};

EXPORT const STRPTR ccstring[2][4] =
{ { "eq",
    "gt",
    "lt",
    "un"
  },
  { "EQ",
    "GT",
    "LT",
    "  "
} }, cctrue[4] =
{   "if == ",
    "if > ",
    "if < ",
    ""
}, ccfalse[4] =
{   "if != ",
    "if <= ",
    "if >= ",
    "if 0 "
};

// IMPORTED VARIABLES-----------------------------------------------------

IMPORT       int                  connected,
                                  consoleopen,
                                  logeprom,
                                  loginstructions,
                                  logbios,
                                  logsubroutines,
                                  logreadwrites,
                                  machine,
                                  memmap,
                                  malzak_x,
                                  pauselog,
                                  pvibase,
                                  rast,
                                  recmode,
                                  runtointerrupt,
                                  runtoloopend,
                                  s_intdir,
                                  s_io,
                                  step,
                                  style,
                                  supercpu,
                                  trace,
                                  traceverbose,
                                  watchreads;
IMPORT       TEXT                 addrstr[32 + 1 + 1],
                                  cd2650_chars[128 + 1],
                                  friendly[21 + 1];
IMPORT       UBYTE                g_bank[1024],
                                  s_glow,
                                  s_toggles,
                                  m_field[16][16];
IMPORT       UWORD                console[4],
                                  jf[2],
                                  mirror_r[32768],
                                  mirror_w[32768];
IMPORT       ULONG                frames,
                                  memflags[32768],
                                  paused,
                                  sound;
IMPORT       FLAG                 belmach,
                                  belmach0,
                                  changed,
                                  life,
                                  piano,
                                  p_morse,
                                  music;
IMPORT const STRPTR               opcodename[256];
IMPORT const struct MachineStruct machines[MACHINES];

#ifdef AMIGA
    IMPORT struct Catalog* CatalogPtr;
    IMPORT BPTR            SpeakHandle;
    IMPORT int             speakoffset,
                           usespeech;
    IMPORT UBYTE           SpeakBuffer[80 + 1];
#endif
#ifdef WIN32
    IMPORT int             CatalogPtr;
    IMPORT HINSTANCE       InstancePtr;
#endif

// MODULE VARIABLES-------------------------------------------------------

MODULE SWORD  res;
MODULE UBYTE  operand,
              before,
              c;
MODULE char   an[80 + 1],
              byte1[2 + 1],
              byte2[2 + 1],
              byte3[2 + 1],
              calm[2 + 1] = "##",
              v[2 + 1],
              reg[1 + 1] = "#", // does not include the 'r'
              mn[80 + 1],
              addressstring[34 + 1]; // "*SPRITECOLLIDE,r3+ [SPRITECOLLIDE]" is largest possible
MODULE int    disassembling = FALSE,
              addr;

// EXPORTED VARIABLES-----------------------------------------------------

EXPORT UBYTE  psu,
              psl,
              memory[32768],
              opcode,
              rr, // register (0..6)
              r[7]; // register is a reserved word
EXPORT UWORD  iar,
              ras[8];
EXPORT ULONG  cycles,
              ea,
              elapsed;
EXPORT int    zinterrupt,
              overcalc, // must be signed!
              slice;
EXPORT char   psubits[3 + 1],
              pslbits[6 + 1];

#ifdef PROFILE
    EXPORT ULONG           profile[256];
#endif
MODULE __inline void set_cc(UBYTE n);
MODULE __inline void pushras(void);
MODULE __inline void pullras(void);
MODULE __inline void branch(void);
MODULE void compare(UBYTE first, UBYTE second);
MODULE UBYTE add(UBYTE dest, UBYTE source);
MODULE UBYTE subtract(UBYTE dest, UBYTE source);
MODULE void ABS_EA(void);
MODULE void BRA_EA(void);
MODULE void REL_EA(void);
MODULE void ZERO_EA(void);
MODULE UBYTE cpuread(int address);
MODULE __inline void cpuwrite(int address, UBYTE data);
MODULE void do_runtoloopend(void);

// these are only needed for debugging
MODULE __inline void tracecpu(ULONG cycles);
MODULE void ABS_EA_TRACE(void);
MODULE void REL_EA_TRACE(void);
MODULE void BRA_EA_TRACE(void);
MODULE void ZERO_EA_TRACE(void);

EXPORT void cpu(void)
{   // all PERSISTent for speed
    PERSIST int  i,
                 sp;

    if (slice <= 0) // this can happen! remember some instructions take up to 18 cycles!
    {   overcalc = -slice;
        return;
    }

    if (zinterrupt)
    {   if (trace || runtointerrupt || logsubroutines)
        {   T1; textpen(TEXTPEN_TRACE);

            if (logsubroutines)
            {   // if psu & PSU_SP is used instead of sp, this doesn't indent properly! (Visual C++ bug?)
                sp = psu & PSU_SP;
                // zprintf("%d: ", sp);
                if (sp)
                {   for (i = 0; i < sp; i++)
                    {   zprintf(" ");
            }   }   }

            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_CPU_INTERRUPT,
                    "Generated interrupt at rastline %d (clock %u)!\n\n"
                ),
                rast,
                cycles + elapsed
            );
            T2;

            if (runtointerrupt)
            {   runto();
        }   }

        psu |= PSU_II; // set II bit (PSU |= %00100000)
#ifdef DEBUGINTERRUPTS
        T1;
        zprintf("Interrupt handler (rastline %d): interrupts are now disabled.\n", rast);
        T2;
#endif
        ras[psu & PSU_SP] = iar;
        switch (machine)
        {
        case INTERTON:
        case ELEKTOR:
        case MALZAK:
        case ASTROWARS:
        case GALAXIA:
            ea = 0x0003;
        acase INSTRUCTOR:
            if (s_intdir == INTDIR_DIRECT)
            {   ea = 0x0007;
            } else
            {   // assert(s_intdir == INTERRUPTS_INDIRECT);
                /* Is it allowed for the address at $0007..$0008 to refer
                to pages 1..3, or only page 0?
                We don't allow it, but perhaps it is allowed:
                "This instruction causes the processor to clear address
                bits #13 and #14, the page address bits." But at what
                point are they cleared? */
                memflags[7] |= COVERAGE_ADDRESS;
                // "Interrupts always drive the processor to page zero." - S2650 CPU manual
                ea = ((memory[7] * 256) + memory[8]) & NONPAGE;
                // we should probably direct this through the mirrorring subsystem
            }
     /* adefault: interrupts aren't supported (nor used?) for ARCADIA, PIPBUG, CD2650, CHAOS, DOLPHIN
            ; */
        }

        // increment SP
        if ((psu & PSU_SP) == 7) // overflow
        {   psu &= ~(PSU_SP); // set SP to 0

            if (loginstructions)
            {   T1; textpen(TEXTPEN_LOG);
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_CPU_INTERRUPTOVERFLOW,
                        "Interrupt has caused a stack overflow!\n\n"
                )   );
                T2;
                if (pauselog)
                {   runto();
        }   }   }
        else
        {   psu++;
        }
        branch();
        if (machine == INSTRUCTOR && s_intdir == INTDIR_INDIRECT)
        {   FIFTEEN_CYCLES;
        } else
        {   NINE_CYCLES;
        }

        zinterrupt = FALSE;
    }
/* Interrupt timing is late. Interrupts should presumably occur (ie. the
   branch should actually happen) before the next executed instruction
   after the interrupt flag has been set. Our granularity is not that
   fine. */

    do
    {   if (memflags[iar] & STEPPOINT)
        {   memflags[iar] &= ~(STEPPOINT);

            T1; textpen(TEXTPEN_DEBUG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_CPU_REACHEDSUBEND,
                    "Reached end of subroutine or interrupt and returned to $%X.\n\n"
                ),
                (unsigned int) iar // should use friendly instead
            );
            T2;
            runto();
        }

        if ((memflags[iar] & BREAKPOINT) && !step)
        {   getfriendly(iar);
            T1; textpen(TEXTPEN_DEBUG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_CPU_HITBP,
                    "Hit code breakpoint at %s!\n\n"
                ),
                friendly
            );
            T2;
            runto();
        }

        /* Opcodes and operands should probably be fetched through
           cpuread(), because of mirroring, read/write status, etc.
           This would be slow, and would cause problems with
           watchpoints and read-once addresses.

           Quick (to code) and dirty method of supporting execution
           mirroring. Note that this changes the actual iar,
           which is not how it should be done. Also note that operands
           are not always mirrored correctly in some edge cases. Eg.
           for Arcadia types "G" and "H":

           $2FFF..$3000: lodi,r0 $10

           it would get its operand from $1000 (instead of $3000) on the
           real machine, but gets it from $3000 on the emulator.

           A more elegant and accurate (although slower) way would be
           to just run every access to iar by the CPU through
           mirror_r(). */

        iar    = mirror_r[iar];
        opcode = memory[iar];

#ifdef PROFILE
        profile[opcode]++;
#endif

        memflags[iar] |= COVERAGE_EXECUTE;
        if (table_sz[supercpu][opcode] >= 2)
        {   memflags[WRAP(1)] |= COVERAGE_EXECUTE;
            if (table_sz[supercpu][opcode] == 3)
            {   memflags[WRAP(2)] |= COVERAGE_EXECUTE;
        }   }

        if (trace || step)
        {   tracecpu(elapsed + cycles);
            T1; textpen(TEXTPEN_TRACE);
            if (mn[0])
            {   zprintf("%s\n", mn);
                if (traceverbose)
                {   zprintf(";%s;\n", an);
            }   }
            /* else zprintf("Empty message!\n"); (should not occur) */
            T2;
        }

        switch (opcode)
        {
        case 0:                                                 // LODZ r0
            if (loginstructions)
            {   getfriendly(iar);
                T1; textpen(TEXTPEN_LOG);
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_CPU_INDETERMINATE,
                        "Instruction at %s executed indeterminate opcode $00 (LODZ r0)!\n\n"
                    ),
                    friendly
                );
                T2;
                if (pauselog)
                {   runto();
            }   }
            // hopefully this is the same behaviour as on a genuine console
            set_cc(r[0]);
            ONE_BYTE;
            REG_CYCLES;
        acase 1:                                                // LODZ r1
        case 2:                                                 // LODZ r2
        case 3:                                                 // LODZ r3
            GET_RR;
            if
            (   opcode  == 1
             && machine == PIPBUG
             && iar     == 0x2A4
            )
            {   r[4] = thechin();
                if (r[4] == 0)
                {   iar = 0x2A3; // $2A3 + 1 == $2A4
            }   }
            r[0] = r[rr];
            set_cc(r[0]);
            ONE_BYTE;
            REG_CYCLES;
        acase 4:                                                // LODI,r0
#ifdef SUGGEST
            if (OPERAND == 0)
            {   getfriendly(iar);
                T1; textpen(TEXTPEN_LOG);
                zprintf("Suggest LODI,r0 $00 (instead of EORZ r0) at %s.\n\n", friendly);
                T2;
            }
#endif
            r[0] = OPERAND;
            set_cc(r[0]);
            TWO_BYTES;
            SIX_CYCLES;
        acase 5:                                                // LODI,r1
        case 6:                                                 // LODI,r2
        case 7:                                                 // LODI,r3
            if
            (   p_morse
             && opcode  == 6 // LODI,r2
             && (   iar == 0x49B
                 || iar == 0x4AC
            )   )
            {   // assert(machine == PIPBUG);
                playsound(FALSE);
            } elif
            (   piano
             && opcode == 5 // LODI,r1
             && iar == 0x1816
            )
            {   // assert(machine == PHUNSY);
                playsound(FALSE);
            }

            GET_RR;
            r[rr] = OPERAND;
            set_cc(r[rr]);
            TWO_BYTES;
            SIX_CYCLES;
        acase 8:                                                // LODR,r0
            REL_EA();
            r[0] = cpuread((int) ea);
            set_cc(r[0]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 9:                                                // LODR,r1
        case 10:                                                // LODR,r2
        case 11:                                                // LODR,r3
            REL_EA();
            GET_RR;
            r[rr] = cpuread((int) ea);
            set_cc(r[rr]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 12:                                               // LODA,r0
        case 13:                                                // LODA,r1
        case 14:                                                // LODA,r2
        case 15:                                                // LODA,r3
            ABS_EA(); // calls GET_RR for us
            r[rr] = cpuread((int) ea);
            set_cc(r[rr]);
            THREE_BYTES;
            TWELVE_CYCLES;
        acase 16:                                               // LPSL addr
            if (supercpu)
            {   ABS_EA();
                psl = cpuread((int) ea);
                THREE_BYTES;
                TWELVE_CYCLES; // guess
            } else
            {   if (loginstructions)
                {   T1; textpen(TEXTPEN_LOG);
                    zprintf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_CPU_UNDEFINED,
                            "Instruction at $%04X executed undefined opcode $%02X!\n\n"
                        ),
                        (int) iar,
                        (int) opcode
                    );
                    T2;
                }
                ONE_BYTE;
                SIX_CYCLES; // guess
            }
        acase 17:                                               // SPSL addr
            if (supercpu)
            {   ABS_EA(); // calls GET_RR for us
                cpuwrite((int) ea, psl);
                THREE_BYTES;
                TWELVE_CYCLES; // guess
            } else
            {   if (loginstructions)
                {   T1; textpen(TEXTPEN_LOG);
                    zprintf
                    (   GetCatalogStr(CatalogPtr,
                            MSG_CPU_UNDEFINED,
                            "Instruction at $%04X executed undefined opcode $%02X!\n\n"
                        ),
                        (int) iar,
                        (int) opcode
                    );
                    T2;
                }
                ONE_BYTE;
                SIX_CYCLES; // guess
            }
        acase 18:                                               // SPSU
            r[0] = psu;
            set_cc(psu);
            ONE_BYTE;
            SIX_CYCLES;
        acase 19:                                               // SPSL
            r[0] = psl;
            set_cc(psl);
            ONE_BYTE;
            SIX_CYCLES;
        acase 20:                                               // RETC,eq
        case 21:                                                // RETC,gt
        case 22:                                                // RETC,lt
            if (BRANCHCODE == CCFIELD)
            {   pullras();
            } else
            {   ONE_BYTE;
            }
            NINE_CYCLES;
        acase 23:                                               // RETC,un
            pullras();
            NINE_CYCLES;
        acase 24:                                               // BCTR,eq
        case 25:                                                // BCTR,gt
        case 26:                                                // BCTR,lt
            if (BRANCHCODE == CCFIELD)
            {   REL_EA();
                ea = mirror_r[ea];
                branch();
            } else
            {   TWO_BYTES;
            }
            NINE_CYCLES;
        acase 27:                                               // BCTR,un
            REL_EA();
            ea = mirror_r[ea];
            branch();
            NINE_CYCLES;
        acase 28:                                               // BCTA,eq
        case 29:                                                // BCTA,gt
        case 30:                                                // BCTA,lt
            if
            (   opcode  == 30
             && music
          // && machine == PIPBUG
             && iar     == 0x4A5
            )
            {   playsound(FALSE);
            }
            if (BRANCHCODE == CCFIELD)
            {   BRA_EA();
                branch();
            } else
            {   THREE_BYTES;
            }
            NINE_CYCLES;
        acase 31:                                               // BCTA,un
            BRA_EA();
            branch();
            NINE_CYCLES;
        acase 32:                                               // EORZ r0
            r[0] = 0;
            psl &= ~(PSL_CC); // clear CC bits of PSL (ie. CC = "eq")
            ONE_BYTE;
            SIX_CYCLES;
        acase 33:                                               // EORZ r1
        case 34:                                                // EORZ r2
        case 35:                                                // EORZ r3
            GET_RR;
            r[0] ^= r[rr];
            set_cc(r[0]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 36:                                               // EORI,r0
            r[0] ^= OPERAND;
            set_cc(r[0]);
            TWO_BYTES;
            SIX_CYCLES;
        acase 37:                                               // EORI,r1
        case 38:                                                // EORI,r2
        case 39:                                                // EORI,r3
            GET_RR;
            r[rr] ^= OPERAND;
            set_cc(r[rr]);
            TWO_BYTES;
            SIX_CYCLES;
        acase 40:                                               // EORR,r0
            REL_EA();
            r[0] ^= cpuread((int) ea);
            set_cc(r[0]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 41:                                               // EORR,r1
        case 42:                                                // EORR,r2
        case 43:                                                // EORR,r3
            REL_EA();
            GET_RR;
            r[rr] ^= cpuread((int) ea);
            set_cc(r[rr]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 44:                                               // EORA,r0
        case 45:                                                // EORA,r1
        case 46:                                                // EORA,r2
        case 47:                                                // EORA,r3
            ABS_EA(); // calls GET_RR for us
            r[rr] ^= cpuread((int) ea);
            set_cc(r[rr]);
            THREE_BYTES;
            TWELVE_CYCLES;
        acase 48:                                               // REDC r0
        case 49:                                                // REDC r1
        case 50:                                                // REDC r2
        case 51:                                                // REDC r3
            GET_RR;
            r[rr] = readport(PORTC);
            set_cc(r[rr]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 52:                                               // RETE,eq
        case 53:                                                // RETE,gt
        case 54:                                                // RETE,lt
            if (loginstructions && machine == ARCADIA)
            {   T1; textpen(TEXTPEN_LOG);
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_CPU_STRANGE,
                        "Instruction at $%X executed strange opcode"
                    ),
                    (long unsigned int) iar
                );
                zprintf
                (   " $%X (RETE,%s)!\n\n",
                    (long unsigned int) opcode,
                                        ccstring[STYLE_SIGNETICS][BRANCHCODE]
                );
                T2;
            }
            if (BRANCHCODE == CCFIELD)
            {   psu &= ~(PSU_II); // clear II bit (PSU &= %11011111)
#ifdef DEBUGINTERRUPTS
                T1;
                zprintf("RETE at $%04X (rastline  %d): interrupts are now enabled.\n", iar, rast);
                T2;
#endif
                pullras();
            } else
            {   ONE_BYTE;
            }
            NINE_CYCLES;
        acase 55:                                               // RETE,un
            if (loginstructions && machine == ARCADIA)
            {   T1; textpen(TEXTPEN_LOG);
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_CPU_STRANGE,
                        "Instruction at $%X executed strange opcode"
                    ),
                    (long unsigned int) iar
                );
                zprintf
                (   " $37 (RETE,un)!\n\n"
                );
                T2;
            }
            psu &= ~(PSU_II); // clear II bit (PSU &= %11011111)
#ifdef DEBUGINTERRUPTS
                T1;
                zprintf("RETE at $%04X (rastline %d): interrupts are now enabled.\n", iar, rast);
                T2;
#endif
            pullras();
            NINE_CYCLES;
        acase 56:                                               // BSTR,eq
        case 57:                                                // BSTR,gt
        case 58:                                                // BSTR,lt
            if (BRANCHCODE == CCFIELD)
            {   REL_EA();
                ea = mirror_r[ea];
                gosub;
            } else
            {   TWO_BYTES;
            }
            NINE_CYCLES;
        acase 59:                                               // BSTR,un
            REL_EA();
            ea = mirror_r[ea];
            gosub;
            NINE_CYCLES;
        acase 60:                                               // BSTA,eq
        case 61:                                                // BSTA,gt
        case 62:                                                // BSTA,lt
        case 63:                                                // BSTA,un
            if (opcode == 63 || BRANCHCODE == CCFIELD)
            {   BRA_EA();
                gosub;
            } else
            {   THREE_BYTES;
            }
            NINE_CYCLES;
        acase 64:                                               // HALT
            if (loginstructions)
            {   getfriendly(iar);
                T1; textpen(TEXTPEN_LOG);
                zprintf("Instruction at %s executed HALT opcode!\n\n", friendly);
                T2;
                if (pauselog)
                {   runto();
            }   }
            /* Strictly speaking, we should maybe halt the emulation
               entirely, instead of just repeatedly executing
               the HALT instruction (since instruction fetching is
               halted). But this way is better, because
               interrupts may still occur. */
            // ZERO_BYTES;
            SIX_CYCLES;
        acase 65:                                               // ANDZ r1
        case 66:                                                // ANDZ r2
        case 67:                                                // ANDZ r3
            GET_RR;
            r[0] &= r[rr];
            set_cc(r[0]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 68:                                               // ANDI,r0
            r[0] &= OPERAND;
            set_cc(r[0]);
            TWO_BYTES;
            SIX_CYCLES;
        acase 69:                                               // ANDI,r1
        case 70:                                                // ANDI,r2
        case 71:                                                // ANDI,r3
            GET_RR;
            r[rr] &= OPERAND;
            set_cc(r[rr]);
            TWO_BYTES;
            SIX_CYCLES;
        acase 72:                                               // ANDR,r0
            REL_EA();
            r[0] &= cpuread((int) ea);
            set_cc(r[0]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 73:                                               // ANDR,r1
        case 74:                                                // ANDR,r2
        case 75:                                                // ANDR,r3
            REL_EA();
            GET_RR;
            r[rr] &= cpuread((int) ea);
            set_cc(r[rr]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 76:                                               // ANDA,r0
        case 77:                                                // ANDA,r1
        case 78:                                                // ANDA,r2
        case 79:                                                // ANDA,r3
            ABS_EA(); // calls GET_RR for us
            r[rr] &= cpuread((int) ea);
            set_cc(r[rr]);
            THREE_BYTES;
            TWELVE_CYCLES;
        acase 80:                                               // RRR,r0
        case 81:                                                // RRR,r1
        case 82:                                                // RRR,r2
        case 83:                                                // RRR,r3
            GET_RR;
            before = r[rr];
            if (psl & PSL_WC)
            {   c = psl & PSL_C;
                psl &= ~(PSL_C + PSL_IDC);
                r[rr] = (before >> 1) | (c << 7);
                psl |= (before & PSL_C) + (r[rr] & PSL_IDC);
            } else
            {   r[rr] = (before >> 1) | (before << 7);
            }
            psl &= ~(PSL_CC | PSL_OVF);
            if (r[rr] >= 128)
            {   psl |= 0x80; // CC = %10
            } elif (r[rr] > 0)
            {   psl |= 0x40; // CC = %01
            }
            if ((before & 0x80) != (r[rr] & 0x80))
            {   psl |= PSL_OVF;
            }
            ONE_BYTE;
            SIX_CYCLES;
        acase 84:                                               // REDE,r0
        case 85:                                                // REDE,r1
        case 86:                                                // REDE,r2
        case 87:                                                // REDE,r3
            GET_RR;
            r[rr] = readport(OPERAND);
            set_cc(r[rr]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 88:                                               // BRNR,r0
        case 89:                                                // BRNR,r1
        case 90:                                                // BRNR,r2
        case 91:                                                // BRNR,r3
            GET_RR;
            if (r[rr])
            {   REL_EA();
                ea = mirror_r[ea];
                branch();
                NINE_CYCLES;
            } else
            {   TWO_BYTES;
                NINE_CYCLES;
                if (runtoloopend)
                {   do_runtoloopend();
            }   }
        acase 92:                                               // BRNA,r0
        case 93:                                                // BRNA,r1
        case 94:                                                // BRNA,r2
        case 95:                                                // BRNA,r3
            GET_RR;
            if (r[rr])
            {   BRA_EA();
                branch();
                NINE_CYCLES;
            } else
            {   THREE_BYTES;
                NINE_CYCLES;
                if (runtoloopend)
                {   do_runtoloopend();
            }   }
        acase 96:                                               // IORZ r0
            // r[0] |= r[0]; is not needed
            set_cc(r[0]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 97:                                               // IORZ r1
        case 98:                                                // IORZ r2
        case 99:                                                // IORZ r3
            GET_RR;
            r[0] |= r[rr];
            set_cc(r[0]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 100:                                              // IORI,r0
            r[0] |= OPERAND;
            set_cc(r[0]);
            TWO_BYTES;
            SIX_CYCLES;
        acase 101:                                              // IORI,r1
        case 102:                                               // IORI,r2
        case 103:                                               // IORI,r3
            GET_RR;
            r[rr] |= OPERAND;
            set_cc(r[rr]);
            TWO_BYTES;
            SIX_CYCLES;
        acase 104:                                              // IORR,r0
            REL_EA();
            r[0] |= cpuread((int) ea);
            set_cc(r[0]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 105:                                              // IORR,r1
        case 106:                                               // IORR,r2
        case 107:                                               // IORR,r3
            REL_EA();
            GET_RR;
            r[rr] |= cpuread((int) ea);
            set_cc(r[rr]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 108:                                              // IORA,r0
        case 109:                                               // IORA,r1
        case 110:                                               // IORA,r2
        case 111:                                               // IORA,r3
            ABS_EA(); // calls GET_RR for us
            r[rr] |= cpuread((int) ea);
            set_cc(r[rr]);
            THREE_BYTES;
            TWELVE_CYCLES;
        acase 112:                                              // REDD r0
        case 113:                                               // REDD r1
        case 114:                                               // REDD r2
        case 115:                                               // REDD r3
            GET_RR;
            r[rr] = readport(PORTD);
            set_cc(r[rr]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 116:                                              // CPSU
            psu &= (~(OPERAND & PSU_WRITABLE));
#ifdef DEBUGINTERRUPTS
            T1;
            if (psu & PSU_II)
            {   zprintf("CPSU at $%04X (rastline %d): interrupts are now disabled.\n", iar, rast);
            } else
            {   zprintf("CPSU at $%04X (rastline %d): interrupts are now enabled.\n", iar, rast);
            }
            T2;
#endif
            TWO_BYTES;
            NINE_CYCLES;
        acase 117:                                              // CPSL
            psl &= (~OPERAND);
            TWO_BYTES;
            NINE_CYCLES;
        acase 118:                                              // PPSU
            psu |= (OPERAND & PSU_WRITABLE);
#ifdef DEBUGINTERRUPTS
            T1;
            if (psu & PSU_II)
            {   zprintf("PPSU at $%04X (rastline %d): interrupts are now disabled.\n", iar, rast);
            } else
            {   zprintf("PPSU at $%04X (rastline %d): interrupts are now enabled.\n", iar, rast);
            }
#endif
            TWO_BYTES;
            NINE_CYCLES;
        acase 119:                                              // PPSL
            psl |= OPERAND;
            if
            (   machine == PIPBUG
             && (            iar == 0x2B4  // COUT
                 || (life && iar == 0xC39) // LIFECOUT
            )   )
            {   ph_cout(FALSE, r[0]);
                // we also let the real COUT/LIFECOUT run, as it has important side effects
            }
            TWO_BYTES;
            NINE_CYCLES;
        acase 120:                                              // BSNR,r0
            if (r[0])
            {   REL_EA();
                ea = mirror_r[ea];
                gosub;
            } else
            {   TWO_BYTES;
            }
            NINE_CYCLES;
        acase 121:                                              // BSNR,r1
        case 122:                                               // BSNR,r2
        case 123:                                               // BSNR,r3
            GET_RR;
            if (r[rr])
            {   REL_EA();
                ea = mirror_r[ea];
                gosub;
            } else
            {   TWO_BYTES;
            }
            NINE_CYCLES;
        acase 124:                                              // BSNA,r0
            if (r[0])
            {   BRA_EA();
                gosub;
            } else
            {   THREE_BYTES;
            }
            NINE_CYCLES;
        acase 125:                                              // BSNA,r1
        case 126:                                               // BSNA,r2
        case 127:                                               // BSNA,r3
            GET_RR;
            if (r[rr])
            {   BRA_EA();
                gosub;
            } else
            {   THREE_BYTES;
            }
            NINE_CYCLES;
        acase 128:                                              // ADDZ r0
            r[0] = add(r[0], r[0]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 129:                                              // ADDZ r1
        case 130:                                               // ADDZ r2
        case 131:                                               // ADDZ r3
            GET_RR;
            r[0] = add(r[0], r[rr]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 132:                                              // ADDI,r0
            r[0] = add(r[0], OPERAND);
            TWO_BYTES;
            SIX_CYCLES;
        acase 133:                                              // ADDI,r1
        case 134:                                               // ADDI,r2
        case 135:                                               // ADDI,r3
            GET_RR;
            r[rr] = add(r[rr], OPERAND);
            TWO_BYTES;
            SIX_CYCLES;
        acase 136:                                              // ADDR,r0
            REL_EA();
            r[0] = add(r[0], cpuread((int) ea));
            TWO_BYTES;
            NINE_CYCLES;
        acase 137:                                              // ADDR,r1
        case 138:                                               // ADDR,r2
        case 139:                                               // ADDR,r3
            REL_EA();
            GET_RR;
            r[rr] = add(r[rr], cpuread((int) ea));
            TWO_BYTES;
            NINE_CYCLES;
        acase 140:                                              // ADDA,r0
        case 141:                                               // ADDA,r1
        case 142:                                               // ADDA,r2
        case 143:                                               // ADDA,r3
            ABS_EA(); // calls GET_RR for us
            r[rr] = add(r[rr], cpuread((int) ea));
            THREE_BYTES;
            TWELVE_CYCLES;
        acase 146:                                              // LPSU
            psu &= PSU_READONLY;
            psu |= (r[0] & PSU_WRITABLE);
#ifdef DEBUGINTERRUPTS
            T1;
            if (psu & PSU_II)
            {   zprintf("LPSU at $%04X (rastline %d): interrupts are now disabled.\n", iar, rast);
            } else
            {   zprintf("LPSU at $%04X (rastline %d): interrupts are now enabled.\n", iar, rast);
            }
            T2;
#endif
            ONE_BYTE;
            SIX_CYCLES;
        acase 147:                                              // LPSL
            psl = r[0];
            ONE_BYTE;
            SIX_CYCLES;
        acase 148:                                              // DAR  r0
            if ((psl & PSL_C) == 0)
            {   // add ten to the tens digit (the high nybble)
                r[0] += 0xA0; // 160. overflow OK. %1010,0000
            }
            if ((psl & PSL_IDC) == 0)
            {   // OR the high nybble with the (low nybble + 10)
                r[0] = (r[0] & 0xF0) | ((r[0] + 0x0A) & 0x0F);
            }
            set_cc(r[0]);
            ONE_BYTE;
            NINE_CYCLES;
        acase 149:                                              // DAR  r1
        case 150:                                               // DAR  r2
        case 151:                                               // DAR  r3
            GET_RR;
            if ((psl & PSL_C) == 0)
            {   // add ten to the tens digit (the high nybble)
                r[rr] += 0xA0; // 160. overflow OK. %1010,0000
            }
            if ((psl & PSL_IDC) == 0)
            {   // OR the high nybble with the (low nybble + 10)
                r[rr] = (r[rr] & 0xF0) | ((r[rr] + 0x0A) & 0x0F);
            }
            set_cc(r[rr]);
            ONE_BYTE;
            NINE_CYCLES;
        acase 152:                                              // BCFR,eq
        case 153:                                               // BCFR,gt
        case 154:                                               // BCFR,lt
            if (BRANCHCODE != CCFIELD)
            {   REL_EA();
                ea = mirror_r[ea];
                branch();
            } else
            {   TWO_BYTES;
            }
            NINE_CYCLES;
        acase 155:                                              // ZBRR
            ZERO_EA();
            branch();
            NINE_CYCLES;
        acase 156:                                              // BCFA,eq
        case 157:                                               // BCFA,gt
        case 158:                                               // BCFA,lt
            if (BRANCHCODE != CCFIELD)
            {   BRA_EA();
                branch();
            } else
            {   THREE_BYTES;
            }
            NINE_CYCLES;
        acase 159:                                              // BXA,r3
            BRA_EA();
            if (psl & PSL_RS)
            {   ea = (ea + r[6]) & AMSK;
            } else
            {   ea = (ea + r[3]) & AMSK;
            }
            branch();
            NINE_CYCLES;
        acase 160:                                              // SUBZ r0
            r[0] = subtract(r[0], r[0]); // not just r[0] = 0; because we need to set PSW bits
            ONE_BYTE;
            SIX_CYCLES;
        acase 161:                                              // SUBZ r1
        case 162:                                               // SUBZ r2
        case 163:                                               // SUBZ r3
            GET_RR;
            r[0] = subtract(r[0], r[rr]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 164:                                              // SUBI,r0
            r[0] = subtract(r[0], OPERAND);
            TWO_BYTES;
            SIX_CYCLES;
        acase 165:                                              // SUBI,r1
        case 166:                                               // SUBI,r2
        case 167:                                               // SUBI,r3
            GET_RR;
            r[rr] = subtract(r[rr], OPERAND);
            TWO_BYTES;
            SIX_CYCLES;
        acase 168:                                              // SUBR,r0
            REL_EA();
            r[0] = subtract(r[0], cpuread((int) ea));
            TWO_BYTES;
            NINE_CYCLES;
        acase 169:                                              // SUBR,r1
        case 170:                                               // SUBR,r2
        case 171:                                               // SUBR,r3
            REL_EA();
            GET_RR;
            r[rr] = subtract(r[rr], cpuread((int) ea));
            TWO_BYTES;
            NINE_CYCLES;
        acase 172:                                              // SUBA,r0
        case 173:                                               // SUBA,r1
        case 174:                                               // SUBA,r2
        case 175:                                               // SUBA,r3
            ABS_EA(); // this calls GET_RR for us
            r[rr] = subtract(r[rr], cpuread((int) ea));
            THREE_BYTES;
            TWELVE_CYCLES;
        acase 176:                                              // WRTC r0
        case 177:                                               // WRTC r1
        case 178:                                               // WRTC r2
        case 179:                                               // WRTC r3
            GET_RR;
            writeport(PORTC, r[rr]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 180:                                              // TPSU
            psl &= 0x3F; // set CC to %00xxxxxx ("eq")
            if ((psu & OPERAND) < OPERAND)
            {   psl |= 0x80; // set CC to %10xxxxxx ("lt")
            }
            TWO_BYTES;
            NINE_CYCLES;
        acase 181:                                              // TPSL
            psl &= 0x3F; // set CC to %00xxxxxx ("eq")
            if ((psl & OPERAND) < OPERAND)
            {   psl |= 0x80; // set CC to %10xxxxxx ("lt")
            }
            TWO_BYTES;
            NINE_CYCLES;
        acase 184:                                              // BSFR,eq
        case 185:                                               // BSFR,gt
        case 186:                                               // BSFR,lt
            if (BRANCHCODE != CCFIELD)
            {   REL_EA();
                ea = mirror_r[ea];
                gosub;
            } else
            {   TWO_BYTES;
            }
            NINE_CYCLES;
        acase 187:                                              // ZBSR
            ZERO_EA();
            // we are making an assumption that the USE monitor switches to the
            // main register bank when these functions are called; this assumption is
            // probably not true.
            if (machine == INSTRUCTOR && logbios)
            {   getfriendly(ea);
                T1; textpen(TEXTPEN_LOG);
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_CPU_LOGMONITOR,
                        "BIOS code at %s was called by game code at $%X!\n\n"
                    ),
                    friendly,
                    (unsigned int) iar
                );
                T2;
            }
            gosub;
            NINE_CYCLES;
        acase 188:                                              // BSFA,eq
        case 189:                                               // BSFA,gt
        case 190:                                               // BSFA,lt
            if (BRANCHCODE != CCFIELD)
            {   BRA_EA();
                gosub;
            } else
            {   THREE_BYTES;
            }
            NINE_CYCLES;
        acase 191:                                              // BSXA,r3
            BRA_EA();
            if (psl & PSL_RS)
            {   ea = (ea + r[6]) & AMSK;
            } else
            {   ea = (ea + r[3]) & AMSK;
            }
            gosub;
            NINE_CYCLES;
        acase 192:                                              // NOP
            ONE_BYTE;
            SIX_CYCLES; // possibly REG_CYCLES;
        acase 193:                                              // STRZ r1
        case 194:                                               // STRZ r2
        case 195:                                               // STRZ r3
            GET_RR;
            r[rr] = r[0];
            set_cc(r[rr]);
            ONE_BYTE;
            REG_CYCLES;
        acase 200:                                              // STRR,r0
            REL_EA();
            cpuwrite((int) ea, r[0]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 201:                                              // STRR,r1
        case 202:                                               // STRR,r2
        case 203:                                               // STRR,r3
            REL_EA();
            GET_RR;
            cpuwrite((int) ea, r[rr]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 204:                                              // STRA,r0
        case 205:                                               // STRA,r1
        case 206:                                               // STRA,r2
        case 207:                                               // STRA,r3
            ABS_EA(); // this calls GET_RR for us
            cpuwrite((int) ea, r[rr]);
            THREE_BYTES;
            TWELVE_CYCLES;
        acase 208:                                              // RRL,r0
        case 209:                                               // RRL,r1
        case 210:                                               // RRL,r2
        case 211:                                               // RRL,r3
            GET_RR;
            if
            (   machine == PHUNSY
             && opcode == 208 // rrl,r0
             && (   (belmach  && iar == 0x182A)
                 || (belmach0 && iar == 0x303A)
            )   )
            {   playsound(FALSE);
            }

            before = r[rr];
            if (psl & PSL_WC)
            {   c = psl & PSL_C;
                psl &= ~(PSL_C + PSL_IDC);
                r[rr] = (before << 1) | c;
                psl |= (before >> 7) + (r[rr] & PSL_IDC);
            } else
            {   r[rr] = (before << 1) | (before >> 7);
            }
            psl &= ~(PSL_CC | PSL_OVF);
            if (r[rr] >= 128)
            {   psl |= 0x80; // CC = %10
            } elif (r[rr] > 0)
            {   psl |= 0x40; // CC = %01
            }
            if ((before & 0x80) != (r[rr] & 0x80))
            {   psl |= PSL_OVF;
            }
            ONE_BYTE;
            SIX_CYCLES;
        acase 212:                                              // WRTE r0
        case 213:                                               // WRTE r1
        case 214:                                               // WRTE r2
        case 215:                                               // WRTE r3
            GET_RR;
            writeport(OPERAND, r[rr]);
            TWO_BYTES;
            NINE_CYCLES;
        acase 216:                                              // BIRR,r0
        case 217:                                               // BIRR,r1
        case 218:                                               // BIRR,r2
        case 219:                                               // BIRR,r3
            GET_RR;
            if (++r[rr]) // overflow OK
            {   REL_EA();
                ea = mirror_r[ea];
                branch();
                NINE_CYCLES;
            } else
            {   TWO_BYTES;
                NINE_CYCLES;
                if (runtoloopend)
                {   do_runtoloopend();
            }   }
        acase 220:                                              // BIRA,r0
        case 221:                                               // BIRA,r1
        case 222:                                               // BIRA,r2
        case 223:                                               // BIRA,r3
            GET_RR;
            if (++r[rr]) // overflow OK
            {   BRA_EA();
                branch();
                NINE_CYCLES;
            } else
            {   THREE_BYTES;
                NINE_CYCLES;
                if (runtoloopend)
                {   do_runtoloopend();
            }   }
        acase 224:                                              // COMZ r0
            psl &= ~(PSL_CC); // clear CC bits of PSL
            ONE_BYTE;
            SIX_CYCLES;
        acase 225:                                              // COMZ r1
        case 226:                                               // COMZ r2
        case 227:                                               // COMZ r3
            GET_RR;
            compare(r[0], r[rr]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 228:                                              // COMI,r0
            compare(r[0], OPERAND);
            TWO_BYTES;
            SIX_CYCLES;
        acase 229:                                              // COMI,r1
        case 230:                                               // COMI,r2
        case 231:                                               // COMI,r3
            GET_RR;
            compare(r[rr], OPERAND);
            TWO_BYTES;
            SIX_CYCLES;
        acase 232:                                              // COMR,r0
            REL_EA();
            compare(r[0], cpuread((int) ea));
            TWO_BYTES;
            NINE_CYCLES;
        acase 233:                                              // COMR,r1
        case 234:                                               // COMR,r2
        case 235:                                               // COMR,r3
            REL_EA();
            GET_RR;
            compare(r[rr], cpuread((int) ea));
            TWO_BYTES;
            NINE_CYCLES;
        acase 236:                                              // COMA,r0
        case 237:                                               // COMA,r1
        case 238:                                               // COMA,r2
        case 239:                                               // COMA,r3
            ABS_EA(); // this calls GET_RR for us
            compare(r[rr], cpuread((int) ea));
            THREE_BYTES;
            TWELVE_CYCLES;
        acase 240:                                              // WRTD r0
        case 241:                                               // WRTD r1
        case 242:                                               // WRTD r2
        case 243:                                               // WRTD r3
            GET_RR;
            writeport(PORTD, r[rr]);
            ONE_BYTE;
            SIX_CYCLES;
        acase 244:                                              // TMI,r0
            psl &= 0x3F; // set CC to %00 ("eq")
            if ((r[0] & OPERAND) < OPERAND)
            {   psl |= 0x80; // set CC to %10 ("lt")
            }
            TWO_BYTES;
            NINE_CYCLES;
        acase 245:                                              // TMI,r1
        case 246:                                               // TMI,r2
        case 247:                                               // TMI,r3
            GET_RR;
            psl &= 0x3F; // set CC to %00 ("eq")
            if ((r[rr] & OPERAND) < OPERAND)
            {   psl |= 0x80; // set CC to %10 ("lt")
            }
            TWO_BYTES;
            NINE_CYCLES;
        acase 248:                                              // BDRR,r0
        case 249:                                               // BDRR,r1
        case 250:                                               // BDRR,r2
        case 251:                                               // BDRR,r3
            GET_RR;
            if (--r[rr]) // underflow OK
            {   REL_EA();
                ea = mirror_r[ea];
                branch();
                NINE_CYCLES;
            } else
            {   TWO_BYTES;
                NINE_CYCLES;
                if (runtoloopend)
                {   do_runtoloopend();
            }   }
        acase 252:                                              // BDRA,r0
        case 253:                                               // BDRA,r1
        case 254:                                               // BDRA,r2
        case 255:                                               // BDRA,r3
            GET_RR;
            if (--r[rr]) // underflow OK
            {   BRA_EA();
                branch();
                NINE_CYCLES;
            } else
            {   THREE_BYTES;
                NINE_CYCLES;
                if (runtoloopend)
                {   do_runtoloopend();
            }   }
        acase 0x90:
        case 0x91:
        case 0xB6:
        case 0xB7:
        case 0xC4:
        case 0xC5:
        case 0xC6:
        case 0xC7:
            if (loginstructions)
            {   T1; textpen(TEXTPEN_LOG);
                zprintf
                (   GetCatalogStr(CatalogPtr,
                        MSG_CPU_UNDEFINED,
                        "Instruction at $%04X executed undefined opcode $%02X!\n\n"
                    ),
                    (int) iar,
                    (int) opcode
                );
                T2;
            }
            // These are all currently defined as 1 byte and 6 cycles
            iar = (iar & PAGE) + ((iar + table_sz[supercpu][opcode]) & NONPAGE);
            cycles += table_clockperiods[supercpu][opcode];
        }

        if (trace || step) // this is too slow!
        {   if (traceverbose)
            {   view_cpu(FALSE);
                if (step)
                {   view_next();
            }   }
            update_monitor(FALSE);
            updatescreen();

            if (step)
            {   step = FALSE;
                memflags[iar] &= ~(STEPPOINT);
                emu_pause(FALSE); // this is recursive! :-(
            }

            // if not pausing, we just checkinput() once
            do
            {   checkinput();
            } while (paused);
    }   }
    while (cycles < (ULONG) slice);

    overcalc = cycles - slice;
    elapsed += cycles;
    cycles = 0; // needed
}

// -----------------------------------------------------------------------

EXPORT void view_cpu(FLAG newline)
{   if (psu & PSU_S  ) psubits[0] = 'S'; else psubits[0] = 's';
    if (psu & PSU_F  ) psubits[1] = 'F'; else psubits[1] = 'f';
    if (psu & PSU_II ) psubits[2] = 'I'; else psubits[2] = 'i';
    psubits[3] = 0;
    if (psl & PSL_IDC) pslbits[0] = 'D'; else pslbits[0] = 'd';
    if (psl & PSL_RS ) pslbits[1] = 'R'; else pslbits[1] = 'r';
    if (psl & PSL_WC ) pslbits[2] = 'W'; else pslbits[2] = 'w';
    if (psl & PSL_OVF) pslbits[3] = 'O'; else pslbits[3] = 'o';
    if (psl & PSL_COM) pslbits[4] = 'M'; else pslbits[4] = 'm';
    if (psl & PSL_C  ) pslbits[5] = 'C'; else pslbits[5] = 'c';
    pslbits[6] = 0;

    T1; textpen(TEXTPEN_VIEW);

    zprintf
    (   "              %7s %10u R0: $%02X R1: $%02X R2: $%02X R3: $%02X PSU: %s\n",
        GetCatalogStr(CatalogPtr, MSG_CLOCK_SHORT,  "Clock:"),
        (unsigned int) (elapsed + cycles),
        r[0],
        r[1],
        r[2],
        r[3],
        psubits
    );
    zprintf
    (   "                                 CC: %s  R4: $%02X R5: $%02X R6: $%02X PSL: %s\n",
        ccstring[STYLE_SIGNETICS][(psl & 0xC0) >> 6], // %11000000 -> %00000011
        r[4],
        r[5],
        r[6],
        pslbits
    );

    zprintf
    (   "                                 SP: %d   $%04X $%04X $%04X $%04X IAR: $%04X\n",
        (int) (psu & PSU_SP),
        ras[0],
        ras[1],
        ras[2],
        ras[3],
        iar
    );

    zprintf
    (   "                                         $%04X $%04X $%04X $%04X\n",
        ras[4],
        ras[5],
        ras[6],
        ras[7]
    );
    if (newline)
    {   zprintf("\n");
    }
    T2;
}

// ** Update condition code reg **
MODULE __inline void set_cc(UBYTE n) // set CC
{   psl &= ~(PSL_CC); // clear CC bits of PSL

    if (n >= 128)
    {   psl |= 0x80; // CC = %10
    } elif (n > 0)
    {   psl |= 0x40; // CC = %01
}   }

// ** Push Return Address Stack **
MODULE __inline void pushras(void)
{   PERSIST int i,  // PERSISTent for speed
                sp; // PERSISTent for speed

    // store address
    ras[psu & PSU_SP] = (iar & PAGE) + ((iar + table_sz[supercpu][opcode]) & NONPAGE);

    if (logsubroutines)
    {   // if psu & PSU_SP is used instead of sp, this doesn't indent properly! (Visual C++ bug?)
        sp = psu & PSU_SP;
        T1; textpen(TEXTPEN_LOG);
        // zprintf("%d: ", sp);
        if (sp)
        {   for (i = 0; i < sp; i++)
            {   zprintf(" ");
        }   }
        getfriendly(ea);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_CPU_CALLING,
                "Calling subroutine %s from $%X...\n\n"
            ),
            friendly,
            iar
        );
        T2;
        if (pauselog)
        {   runto();
    }   }

    // increment SP
    if ((psu & PSU_SP) == 7) // overflow
    {   psu &= ~(PSU_SP); // set SP to 0

        if (loginstructions || logsubroutines)
        {   T1; textpen(TEXTPEN_LOG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_CPU_SUBROUTINEOVERFLOW,
                    "Stack overflow at $%X!\n\n"
                ),
                (unsigned int) iar
            );
            T2;
    }   }
    else
    {   psu++;
}   }

// ** Pull Return Address Stack **
MODULE __inline void pullras(void)
{   PERSIST int i,  // PERSISTent for speed
                sp; // PERSISTent for speed

    // decrement SP
    if ((psu & PSU_SP) == 0) // underflow
    {   psu |= PSU_SP; // set SP to 7

        if (loginstructions || logsubroutines)
        {   T1; textpen(TEXTPEN_LOG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_CPU_UNDERFLOW,
                    "Stack underflow at $%X!\n\n"
                ),
                (unsigned int) iar
            );
            T2;
            if (pauselog)
            {   runto();
    }   }   }
    else psu--;

    if (logsubroutines)
    {   // if psu & PSU_SP is used instead of sp, this doesn't indent properly! (Visual C++ bug?)
        sp = psu & PSU_SP;
        T1; textpen(TEXTPEN_LOG);
        // zprintf("%d: ", sp);
        if (sp)
        {   for (i = 0; i < sp; i++)
            {   zprintf(" ");
        }   }
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_CPU_RETURNING,
                "Returning to $%X from subroutine or interrupt at $%X.\n\n"
            ),
            ras[psu & PSU_SP],
            (unsigned int) iar // should use friendly instead
        );
        T2;
        if (pauselog)
        {   runto();
    }   }

    iar = ras[psu & PSU_SP];
}

MODULE __inline void branch(void)
{   /* page 0 is $0000..$1FFF
       page 1 is $2000..$3FFF
       page 2 is $4000..$5FFF
       page 3 is $6000..$7FFF
    ippaaaaa aaaaaaaa */

    if
    (   logbios
     && (   (   machine == ELEKTOR
             && iar >= 0x800 // we don't log the monitor calling itself
             && ea  <= 0x7FF // this handles page and non-page components
             && !zinterrupt
            )
         || (   (machine == PIPBUG || machine == CD2650)
             && iar >= 0x400 // we don't log the monitor calling itself
             && ea  <= 0x3FF // this handles page and non-page components
          // && !zinterrupt is not necessary
            )
         || (   machine == PHUNSY
             && iar >= 0x800 // we don't log the monitor calling itself
             && ea  <= 0x7FF // this handles page and non-page components
    )   )   )
    {   getfriendly(ea);
        T1; textpen(TEXTPEN_LOG);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_CPU_LOGMONITOR,
                "BIOS code at %s was called by game code at $%X!\n\n"
            ),
            friendly,
            (unsigned int) iar
        );
        T2;
    }
    iar = (UWORD) ea;
    if (operand & 0x80)
    {   cycles += 6;
}   }

MODULE void compare(UBYTE first, UBYTE second)
{   int d;

    psl &= 0x3F; // clear CC
    if (psl & PSL_COM) // unsigned compare
    {   d = (UBYTE) first - (UBYTE) second;
    } else
    {   d = (SBYTE) first - (SBYTE) second;
    }
    if (d < 0 )
    {   psl |= 0x80; // %10 = lt
    } elif (d > 0)
    {   psl |= 0x40; // %01 = gt
}   }

// Add source to destination
// Add (if WC==0) or add with carry (if WC==1)
MODULE UBYTE add(UBYTE dest, UBYTE source)
{   before = dest;
    res = dest + source + ((psl >> 3) & psl & PSL_C);

    psl &= ~(PSL_CC | PSL_IDC | PSL_OVF | PSL_C);
    if (res & 0x100)
    {   psl |= PSL_C;
    }
    dest = res & 0xFF;
    if ((dest & 15) < (before & 15))
    {   psl |= PSL_IDC;
    }
    if (dest >= 128)
    {   psl |= 0x80; // CC = %10
    } elif (dest > 0)
    {   psl |= 0x40; // CC = %01
    }
    if (source <= 127)
    {   if (before <= 127 && dest >= 128)
        {   psl |= PSL_OVF;
    }   }
    else
    {   // assert(source >= 128);
        if (before >= 128 && dest <= 127)
        {   psl |= PSL_OVF;
    }   }

    /* OVF is set if both numbers were positive (<= 127) and the answer is negative (>= 128).
               or if both numbers were negative (>= 128) and the answer is positive (<= 127).
    (ie. if they both had the same sign bit, and result has the opposite sign bit)
    Otherwise, it is cleared. */

    return dest;
}

// Subtract source from destination
// Subtract with borrow if WC flag of PSL is set
MODULE UBYTE subtract(UBYTE dest, UBYTE source)
{   before = dest;
    res = dest - source - ((psl >> 3) & (psl ^ PSL_C) & PSL_C);

    psl &= ~(PSL_CC | PSL_IDC | PSL_OVF | PSL_C);
    if ((res & 0x100) == 0)
    {   psl |= PSL_C;
    }
    dest = res & 0xFF;
    if ((dest & 15) <= (before & 15))
    {   psl |= PSL_IDC;
    }
    if (dest >= 128)
    {   psl |= 0x80; // CC = %10
    } elif (dest > 0)
    {   psl |= 0x40; // CC = %01
    }
    if (source <= 127)
    {   if (before <= 127 && dest >= 128)
        {   psl |= PSL_OVF;
    }   }
    else
    {   // assert(source >= 128);
        if (before >= 128 && dest <= 127)
        {   psl |= PSL_OVF;
    }   }

    /* OVF is set if both numbers were positive (<= 127) and the answer is negative (>= 128).
               or if both numbers were negative (>= 128) and the answer is positive (<= 127).
    (ie. if they both had the same sign bit, and result has the opposite sign bit)
    Otherwise, it is cleared. */

    return dest;
}

/* ABS_EA(), etc. don't use cpuread(), because they simply calculate an
   address (ie. a number), they don't actually look at the contents of
   that address. */

// ** build effective address with absolute addressing (non-branch) **
// called by LODA, EORA, ANDA, IORA, SUBA, ADDA, STRA, COMA
// ABS_EA() uses GET_RR; the others don't!
MODULE void ABS_EA(void)
{   /* build effective address within current 8K page */

    operand = memory[WRAP(1)];
    ea = (iar & PAGE) + (((operand << 8) + memory[WRAP(2)]) & NONPAGE);
    GET_RR;

    if (operand & 0x80) /* indirect addressing? */
    {   addr = (int) mirror_r[ea]; // needed for Interton BOWLING
        memflags[addr] |= COVERAGE_ADDRESS;
        /* build indirect 32K address */
        ea = memory[addr] << 8;
        if ((++addr & NONPAGE) == 0)
        {   addr -= PLEN; /* page wrap */
        }
        ea = (ea + memory[addr]) & AMSK;
        memflags[addr] |= COVERAGE_ADDRESS;
        SIX_CYCLES;
    }

    /* check indexed addressing modes */
    switch (operand & 0x60)
    {
    case 0x20: /* auto increment indexed */
        r[rr]++;
        ea = (ea & PAGE) + ((ea + r[rr]) & NONPAGE);
        rr = 0; /* absolute addressing reg is R0 */
    acase 0x40: /* auto decrement indexed */
        r[rr]--;
        ea = (ea & PAGE) + ((ea + r[rr]) & NONPAGE);
        rr = 0; /* absolute addressing reg is R0 */
    acase 0x60: /* indexed */
        ea = (ea & PAGE) + ((ea + r[rr]) & NONPAGE);
        rr = 0; /* absolute addressing reg is R0 */
 // adefault: // eg. $00 (not indexed)
}   }

// ** build effective address with absolute addressing (branch) **
// called by BCTA, BRNA, BCFA, BIRA, BDRA, BXA, BSXA, BSTA, BSNA, BSFA
MODULE void BRA_EA(void)
{   /* build address in 32K address space */

    operand = memory[WRAP(1)];
    ea = ((operand << 8) + memory[WRAP(2)]) & AMSK;
    ea = mirror_r[ea]; // all branches are considered to be reads

    if (operand & 0x80) // indirect addressing?
    {   addr = (int) ea;
        memflags[addr] |= COVERAGE_ADDRESS;
        /* build indirect 32K address */
        ea = memory[addr] << 8;

        if ((++addr & NONPAGE) == 0)
        {   addr -= PLEN; /* page wrap */
        }
        memflags[addr] |= COVERAGE_ADDRESS;
        ea = (ea + memory[addr]) & AMSK;
        SIX_CYCLES;
    }

#ifdef SUGGEST
    if (ea >= WRAP(-64) && ea <= WRAP(63))
    {   T1; textpen(TEXTPEN_LOG);
        zprintf("Far jump to near address at $%X (jumping to $%X).\n\n", iar, ea);
        T2;
    } elif (ea <= 63 || ea >= 0x1F80)
    {   T1; textpen(TEXTPEN_LOG);
        zprintf("Far jump to zero page address at $%X (jumping to $%X).\n\n", iar, ea);
        T2;
    }
#endif
}

// ** build effective add ress with relative addressing **
// called by LODR, BCTR, BSTR, EORR, ANDR, BRNR, IORR, BSNR, ADDR,
// BCFR, SUBR, BSFR, STRR, BIRR, COMR, BDRR
MODULE void REL_EA(void)
{   /* build effective address within current 8K page */

    operand = memory[WRAP(1)];
    ea = (iar & PAGE) + ((iar + 2 + S2650_relative[operand]) & NONPAGE);

/* Mirroring is not done here.
For relative branch instructions, mirroring is handled by the calling
routine (after calling this routine).
  For relative non-branch instructions, mirroring is handled by cpuread().
*/

    if (operand & 0x80) /* indirect bit set? */
    {   addr = (int) ea;
        memflags[addr] |= COVERAGE_ADDRESS;
        /* build indirect 32K address */
        ea = memory[addr] << 8;
        if ((++addr & NONPAGE) == 0)
        {   addr -= PLEN; /* page wrap */
        }
        memflags[addr] |= COVERAGE_ADDRESS;
        ea = (ea + memory[addr]) & AMSK;
        SIX_CYCLES;
}   }

// ** build effective address with zero relative addressing **
// called by ZBRR, ZBSR
MODULE void ZERO_EA(void)
{   /* build effective address from 0 */

    operand = memory[WRAP(1)];
    ea = (S2650_relative[operand] & NONPAGE);
    // ea = mirror_r[ea]; // all branches are considered to be reads
    // not really necessary, as there is no mirroring done of $0000..$007F
    // nor $7F80..$7FFF.

    if (operand & 0x80) // indirect bit set?
    {   addr = (int) ea;
        memflags[addr] |= COVERAGE_ADDRESS;
        /* build indirect 32K address */
        ea = memory[addr] << 8;
        if ((++addr & NONPAGE) == 0)
        {   addr -= PLEN; /* page wrap */
        }
        memflags[addr] |= COVERAGE_ADDRESS;
        ea = (ea + memory[addr]) & AMSK;
        SIX_CYCLES;
}   }

MODULE void ABS_EA_TRACE(void)
{   ea = (iar & PAGE) + (((operand << 8) + memory[iar + 2]) & NONPAGE);
    // what about mirroring?

    // addressstring must be made BEFORE the main string
    if (operand & 0x80) /* indirect addressing? */
    {   addressstring[0] = '*';
        number_to_friendly(ea, &addressstring[1], FALSE);
    } else
    {   number_to_friendly(ea, &addressstring[0], FALSE);
    }

    /* check indexed addressing modes */
    switch (operand & 0x60)
    {
    case 0x20: /* auto increment indexed */
        strcat(addressstring, ",r");
        reg[0] = '0' + rr;
        strcat(addressstring, reg);
        strcat(addressstring, "+");
        reg[0] = '0';
#ifdef WIN32
        T1; T2; // workaround for a strange compiler bug (gives wrong trace output otherwise)
#endif
    acase 0x40: /* auto decrement indexed */
        strcat(addressstring, ",r");
        reg[0] = '0' + rr;
        strcat(addressstring, reg);
        strcat(addressstring, "-");
        reg[0] = '0';
#ifdef WIN32
        T1; T2; // workaround for a strange compiler bug (gives wrong trace output otherwise)
#endif
    acase 0x60: /* indexed */
        strcat(addressstring, ",r");
        reg[0] = '0' + rr;
        strcat(addressstring, reg);
        reg[0] = '0';
#ifdef WIN32
        T1; T2; // workaround for a strange compiler bug (gives wrong trace output otherwise)
#endif
 // adefault: // non-indexed
    }
// the above workaround is required for both tracing and disassembling

    if (operand & 0x80)
    {   ea = (iar & PAGE) + (((operand << 8) + memory[iar + 2]) & NONPAGE);
        GET_RR;
        addr = (int) ea;
        ea = memory[addr] << 8;
        if ((++addr & NONPAGE) == 0)
        {   addr -= PLEN; /* page wrap */
        }
        ea = (ea + memory[addr]) & AMSK;
    }

    switch (operand & 0x60)
    {
    case 0x20: /* auto increment indexed */
        ea = (ea & PAGE) + ((ea + r[rr] + 1) & NONPAGE);
    acase 0x40: /* auto decrement indexed */
        ea = (ea & PAGE) + ((ea + r[rr] - 1) & NONPAGE);
    acase 0x60: /* indexed */
        ea = (ea & PAGE) + ((ea + r[rr]    ) & NONPAGE);
 // adefault: // eg. $00 (not indexed)
    }

    if
    (   (operand & 0x80) // indirect
     || (operand & 0x60) // indexed
    )
    {   strcat(addressstring, " [");
        number_to_friendly(ea, friendly, FALSE);
        strcat(addressstring, friendly);
        strcat(addressstring, "]");
}   }

MODULE void REL_EA_TRACE(void)
{   ea = (iar & PAGE) + ((iar + 2 + S2650_relative[operand]) & NONPAGE);

    // what about mirroring?

    if (operand & 0x80) /* indirect addressing? */
    {   addressstring[0] = '*';
        number_to_friendly(ea, &addressstring[1], FALSE);

        addr = (int) ea;
        ea = memory[addr] << 8;
        if ((++addr & NONPAGE) == 0)
        {   addr -= PLEN; /* page wrap */
        }
        ea = (ea + memory[addr]) & AMSK;

        strcat(addressstring, " [");
        number_to_friendly(ea, friendly, FALSE);
        strcat(addressstring, friendly);
        strcat(addressstring, "]");
    } else
    {   number_to_friendly(ea, &addressstring[0], FALSE);
}   }

MODULE void BRA_EA_TRACE(void)
{   ea = ((operand << 8) + memory[WRAP(2)]) & AMSK;
    // what about mirroring?

    if (operand & 0x80) // indirect addressing?
    {   addressstring[0] = '*';
        number_to_friendly(ea, &addressstring[1], FALSE);

        addr = (int) ea;
        ea = memory[addr] << 8;
        if ((++addr & NONPAGE) == 0)
        {   addr -= PLEN; /* page wrap */
        }
        ea = (ea + memory[addr]) & AMSK;

        strcat(addressstring, " [");
        number_to_friendly(ea, friendly, FALSE);
        strcat(addressstring, friendly);
        strcat(addressstring, "]");
    } else
    {   number_to_friendly(ea, &addressstring[0], FALSE);
}   }

MODULE void ZERO_EA_TRACE(void)
{   ea = (S2650_relative[operand] & NONPAGE);

    // what about mirroring?

    if (operand & 0x80) // indirect bit set?
    {   addressstring[0] = '*';
        number_to_friendly(ea, &addressstring[1], FALSE);

        addr = (int) ea;
        ea = memory[addr] << 8;
        if ((++addr & NONPAGE) == 0)
        {   addr -= PLEN; /* page wrap */
        }
        ea = (ea + memory[addr]) & AMSK;

        strcat(addressstring, " [");
        number_to_friendly(ea, friendly, FALSE);
        strcat(addressstring, friendly);
        strcat(addressstring, "]");
    } else
    {   number_to_friendly(ea, &addressstring[0], FALSE);
}   }

EXPORT void disassemble(int address1, int address2)
{   ULONG savedcycles = cycles;
    UWORD savediar    = iar;
    UBYTE savedpsl    = psl;

    cycles = 0;
    iar    = address1;
    psl    = 0;

    disassembling = TRUE;
    do
    {   opcode = memory[iar];
        GET_RR;
        tracecpu(cycles);
        zprintf("%s\n", mn);
        if (traceverbose)
        {   zprintf(";%s;\n", an);
        }
        cycles += table_clockperiods[supercpu][opcode];
        iar    += table_sz[supercpu][opcode]; // we deliberately allow it to flow onto the next page
    } while (iar <= address2);
    disassembling = FALSE;

    cycles = savedcycles;
    iar    = savediar;
    psl    = savedpsl;
}

MODULE __inline void tracecpu(ULONG cycles)
{   operand = memory[WRAP(1)];

    GET_RR;
    reg[0] = '0' + rr;

    if (traceverbose && getfriendly(iar))
    {   T1;
        if (!disassembling)
        {   textpen(TEXTPEN_TRACE);
        }
        zprintf("%s:\n", friendly);
        T2;
    }

    decimal(mn, cycles);
    mn[10] = ' ';
    hex2(&mn[11], iar);
    mn[15] = ' ';

/* typical line is:
clock      iar  machine  assembler                                    ;c,s
1234567890 0000 00       EORZ    R0                                   ;1,2
01234567890123456789012345678901234567890123456789012345678901234567890123
where c is cycle count (2, 3, 4 or 6)
*/
    hex1(byte1, opcode);

    if (table_sz[supercpu][opcode] == 3)
    {   hex1(byte2, operand);
        hex1(byte3, memory[WRAP(2)]);
        sprintf(&mn[16], "%s %s %s ", byte1, byte2, byte3);
    } elif (table_sz[supercpu][opcode] == 2)
    {   hex1(byte2, operand);
        sprintf(&mn[16], "%s %s    ", byte1, byte2);
    } else
    {   // assert(table_sz[supercpu][opcode] == 1);
        sprintf(&mn[16], "%s       ", byte1);
    }

    switch(opcode)
    {
    case 0x0: // @0
    case 0x1: // @1
    case 0x2: // @2
    case 0x3: // @3
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "LODZ    r");
            strcat(mn, reg);
            sprintf(an, "r0 = r%s", reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     A,");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "A = %s", calm);
        }
    acase 0x4: // @4
    case 0x5: // @5
    case 0x6: // @6
    case 0x7: // @7
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "LODI,r");
            strcat(mn, reg);
            strcat(mn, " $");
            strcat(mn, v);
            sprintf(an, "r%s = $%s", reg, v);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",#");
            strcat(mn, v);
            sprintf(an, "%s = $%s", calm, v);
        }
    acase 0x08: // @10
    case 0x09: // @11
    case 0x0A: // @12
    case 0x0B: // @13
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "LODR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s = *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s = *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x0C: // @14
    case 0x0D: // @15
    case 0x0E: // @16
    case 0x0F: // @17
        ABS_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "LODA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s = *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s = *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x10: // @20
        if (supercpu)
        {   ABS_EA_TRACE();
            if (style == STYLE_SIGNETICS)
            {   strcat(mn, "LPSL    ");
                strcat(mn, addressstring);
            } else
            {   // assert(style == STYLE_CALM);
                strcat(mn, "LOAD     L,");
                strcat(mn, addressstring);
            }
            sprintf(an, "PSL = *(%s)", addressstring);
        } else
        {   strcat(mn, "?");
            strcpy(an, "?");
        }
    acase 0x11: // @21
        if (supercpu)
        {   ABS_EA_TRACE();
            if (style == STYLE_SIGNETICS)
            {   strcat(mn, "SPSL    ");
                strcat(mn, addressstring);
            } else
            {   // assert(style == STYLE_CALM);
                strcat(mn, "LOAD     ");
                strcat(mn, addressstring);
                strcat(mn, ",L");
            }
            sprintf(an, "*(%s) = PSL", addressstring);
        } else
        {   strcat(mn, "?");
            strcpy(an, "?");
        }
    acase 0x12: // @22
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "SPSU"); // or "SPSU    r0"
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     A,U");
        }
        strcpy(an, "r0 = PSU");
    acase 0x13: // @23
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "SPSL"); // or "SPSL    r0"
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     A,L");
        }
        strcpy(an, "r0 = PSL");
    acase 0x14: // @24
    case 0x15: // @25
    case 0x16: // @26
    case 0x17: // @27
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "RETC,");
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "RET ");
        }
        strcat(mn, ccstring[style][BRANCHCODE]);
        sprintf(an, "%sreturn", cctrue[BRANCHCODE]);
    acase 0x18: // @30
    case 0x19: // @31
    case 0x1A: // @32
    case 0x1B: // @33
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BCTR,");
            strcat(mn, ccstring[style][BRANCHCODE]);
            strcat(mn, " ");
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "JUMP ");
            strcat(mn, ccstring[style][BRANCHCODE]);
            strcat(mn, "  ");
        }
        strcat(mn, addressstring);
        sprintf(an, "%sgoto %s", cctrue[BRANCHCODE], addressstring);
    acase 0x1C: // @34
    case 0x1D: // @35
    case 0x1E: // @36
    case 0x1F: // @37
        BRA_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BCTA,");
            strcat(mn, ccstring[style][BRANCHCODE]);
            strcat(mn, " ");
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "JUMP ");
            strcat(mn, ccstring[style][BRANCHCODE]);
            strcat(mn, "  ");
        }
        strcat(mn, addressstring);
        sprintf(an, "%sgoto %s", cctrue[BRANCHCODE], addressstring);
    acase 0x20: // @40
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "EORZ    r0");
            sprintf(an, "r0 = 0");
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "CLR      A");
            sprintf(an, "A = 0");
        }
    acase 0x21: // @41
    case 0x22: // @42
    case 0x23: // @43
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "EORZ    r");
            strcat(mn, reg);
            sprintf(an, "r0 ^= r%s", reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "XOR      A,");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "A ^= %s", calm);
        }
    acase 0x24: // @44
    case 0x25: // @45
    case 0x26: // @46
    case 0x27: // @47
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "EORI,r");
            strcat(mn, reg);
            strcat(mn, " $");
            strcat(mn, v);
            sprintf(an, "r%s ^= $%s", reg, v);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "XOR      ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",#");
            strcat(mn, v);
            sprintf(an, "%s ^= $%s", calm, v);
        }
    acase 0x28: // @50
    case 0x29: // @51
    case 0x2A: // @52
    case 0x2B: // @53
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "EORR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s ^= *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "XOR      ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s ^= *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x2C: // @54
    case 0x2D: // @55
    case 0x2E: // @56
    case 0x2F: // @57
        ABS_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "EORA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s ^= *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "XOR      ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s ^= *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x30: // @60
    case 0x31: // @61
    case 0x32: // @62
    case 0x33: // @63
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "REDC    r");
            strcat(mn, reg);
            sprintf(an, "r%s = PORTC", reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",$CTRL");
            sprintf(an, "%s = PORTC", calm);
        }
    acase 0x34: // @64
    case 0x35: // @65
    case 0x36: // @66
    case 0x37: // @67
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "RETE,");
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "RETION");
        }
        strcat(mn, ccstring[style][BRANCHCODE]);
        sprintf(an, "%s{ PSU &= ~PSU_II; return; }", cctrue[BRANCHCODE]);
    acase 0x38: // @70
    case 0x39: // @71
    case 0x3A: // @72
    case 0x3B: // @73
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BSTR,");
            strcat(mn, ccstring[style][BRANCHCODE]);
            strcat(mn, " ");
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "CALL ");
            strcat(mn, ccstring[style][BRANCHCODE]);
            strcat(mn, "  ");
        }
        strcat(mn, addressstring);
        sprintf(an, "%sgosub %s", cctrue[BRANCHCODE], addressstring);
    acase 0x3C: // @74
    case 0x3D: // @75
    case 0x3E: // @76
    case 0x3F: // @77
        BRA_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BSTA,");
            strcat(mn, ccstring[style][BRANCHCODE]);
            strcat(mn, " ");
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "CALL ");
            strcat(mn, ccstring[style][BRANCHCODE]);
            strcat(mn, "  ");
        }
        strcat(mn, addressstring);
        sprintf(an, "%sgosub %s", cctrue[BRANCHCODE], addressstring);
    acase 0x40: // @100
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "HALT");
        } else
        {   strcat(mn, "WAIT");
        }
        strcpy(an, "for (;;)");
    acase 0x41: // @101
    case 0x42: // @102
    case 0x43: // @103
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "ANDZ    r");
            strcat(mn, reg);
            sprintf(an, "r0 &= r%s", reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "AND      A,");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "A &= %s", calm);
        }
    acase 0x44: // @104
    case 0x45: // @105
    case 0x46: // @106
    case 0x47: // @107
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "ANDI,r");
            strcat(mn, reg);
            strcat(mn, " $");
            strcat(mn, v);
            sprintf(an, "r%s &= $%s", reg, v);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "AND      ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",#");
            strcat(mn, v);
            sprintf(an, "%s &= $%s", calm, v);
        }
    acase 0x48: // @110
    case 0x49: // @111
    case 0x4A: // @112
    case 0x4B: // @113
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "ANDR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s &= *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "AND      ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s &= *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x4C: // @114
    case 0x4D: // @115
    case 0x4E: // @116
    case 0x4F: // @117
        ABS_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "ANDA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s &= *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "AND      ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s &= *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x50: // @120
    case 0x51: // @121
    case 0x52: // @122
    case 0x53: // @123
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "RRR,r");
            strcat(mn, reg);
            sprintf(an, "r%s >>= 1", reg);
        } else
        {   // assert(style == STYLE_CALM);
            if (psl & PSL_WC)
            {   strcat(mn, "RRC      ");
            } else
            {   strcat(mn, "RR       ");
            }
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "%s >>= 1", calm);
        }
    acase 0x54: // @124
    case 0x55: // @125
    case 0x56: // @126
    case 0x57: // @127
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "REDE,r");
            strcat(mn, reg);
            strcat(mn, " $");
            strcat(mn, v);
            sprintf(an, "r%s = IOPORT($%s)", reg, v);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",$");
            strcat(mn, v);
            sprintf(an, "%s = IOPORT($%s)", calm, v);
        }
    acase 0x58: // @130
    case 0x59: // @131
    case 0x5A: // @132
    case 0x5B: // @133
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BRNR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "if (r%s != 0) goto %s", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "JUMP ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, "NE ");
            sprintf(an, "if (%s != 0) goto %s", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x5C: // @134
    case 0x5D: // @135
    case 0x5E: // @136
    case 0x5F: // @137
        BRA_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BRNA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "if (r%s != 0) goto %s", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "JUMP ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, "NE ");
            sprintf(an, "if (%s != 0) goto %s", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x60: // @140
    case 0x61: // @141
    case 0x62: // @142
    case 0x63: // @143
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "IORZ    r");
            strcat(mn, reg);
            sprintf(an, "r0 |= r%s", reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "OR       A,");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "A |= %s", calm);
        }
    acase 0x64: // @144
    case 0x65: // @145
    case 0x66: // @146
    case 0x67: // @147
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "IORI,r");
            strcat(mn, reg);
            strcat(mn, " $");
            strcat(mn, v);
            sprintf(an, "r%s |= $%s", reg, v);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "OR       ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",#");
            strcat(mn, v);
            sprintf(an, "%s |= $%s", calm, v);
        }
    acase 0x68: // @150
    case 0x69: // @151
    case 0x6A: // @152
    case 0x6B: // @153
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "IORR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s |= *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "OR       ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s |= *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x6C: // @154
    case 0x6D: // @155
    case 0x6E: // @156
    case 0x6F: // @157
        ABS_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "IORA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s |= *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "OR       ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s |= *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x70: // @160
    case 0x71: // @161
    case 0x72: // @162
    case 0x73: // @163
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "REDD    r");
            strcat(mn, reg);
            sprintf(an, "r%s = PORTD", reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",$DATA");
            sprintf(an, "%s = PORTD", calm);
        }
    acase 0x74: // @164
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "CPSU    $");
            strcat(mn, v);
        } else
        {   // assert(style == STYLE_CALM);
            switch (operand)
            {
            case 0x07: // @7
                strcat(mn, "CLR      STACK");
            acase 0x20: // @40
                strcat(mn, "ION      CLR IOF");
            acase 0x40: // @100
                strcat(mn, "CLR      OUTPUT");
            adefault:
                strcat(mn, "BIC      U,#");
                strcat(mn, v);
        }   }
        sprintf(an, "PSU &= ~($%s & %%01100111)", v);
    acase 0x75: // @165
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "CPSL    $");
            strcat(mn, v);
        } else
        {   // assert(style == STYLE_CALM);
            switch (operand)
            {
            case 0x01: // @1
                strcat(mn, "CLRC     CLR CARRY");
            acase 0x02: // @2
                strcat(mn, "CLR      LOGICOMP");
            acase 0x04: // @4
                strcat(mn, "CLRV     CLR OVERFLOW");
            acase 0x08: // @10
                strcat(mn, "CLR      WITHCARRY");
            acase 0x10: // @20
                strcat(mn, "CLR      BANK"); // BANK1 is acceptable as a synonym of BANK
            adefault:
                strcat(mn, "BIC      L,#");
                strcat(mn, v);
        }   }
        sprintf(an, "PSL &= ~($%s)", v);
    acase 0x76: // @166
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "PPSU    $");
            strcat(mn, v);
        } else
        {   // assert(style == STYLE_CALM);
            switch (operand)
            {
            case 0x20: // @40
                strcat(mn, "IOF      SET IOF");
            acase 0x40: // @100
                strcat(mn, "SET      OUTPUT");
            adefault:
                strcat(mn, "OR       U,#");
                strcat(mn, v);
        }   }
        sprintf(an, "PSU |= $%s & %%01100111", v);
    acase 0x77:  // @167
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "PPSL    $");
            strcat(mn, v);
        } else
        {   // assert(style == STYLE_CALM);
            switch (operand)
            {
            case 0x01: // @1
                strcat(mn, "SETC     SET CARRY");
            acase 0x02: // @2
                strcat(mn, "SET      LOGICOMP");
            acase 0x04: // @4
                strcat(mn, "SETV     SET OVERFLOW");
            acase 0x08: // @10
                strcat(mn, "SET      WITHCARRY");
            acase 0x10: // @20
                strcat(mn, "SET      BANK"); // BANK1 is acceptable as a synonym of BANK
            adefault:
                strcat(mn, "OR       L,#");
                strcat(mn, v);
        }   }
        sprintf(an, "PSL |= $%s", v);
    acase 0x78: // @170
    case 0x79: // @171
    case 0x7A: // @172
    case 0x7B: // @173
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BSNR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "if (r%s != 0) gosub %s", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "CALL ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, "NE ");
            sprintf(an, "if (%s != 0) gosub %s", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x7C: // @174
    case 0x7D: // @175
    case 0x7E: // @176
    case 0x7F: // @177
        BRA_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BSNA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "if (r%s != 0) gosub %s", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "CALL ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, "NE ");
            sprintf(an, "if (%s != 0) gosub %s", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x80: // @200
    case 0x81: // @201
    case 0x82: // @202
    case 0x83: // @203
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "ADDZ    r");
            strcat(mn, reg);
            sprintf(an, "r0 += r%s", reg);
        } else
        {   // assert(style == STYLE_CALM);
            if (psl & PSL_WC)
            {   strcat(mn, "ADDC     A,");
            } else
            {   strcat(mn, "ADD      A,");
            }
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "A += %s", calm);
        }
    acase 0x84: // @204
    case 0x85: // @205
    case 0x86: // @206
    case 0x87: // @207
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "ADDI,r");
            strcat(mn, reg);
            strcat(mn, " $");
            strcat(mn, v);
            if (operand == 1)
            {   sprintf(an, "r%s++", reg);
            } else
            {   sprintf(an, "r%s += $%s", reg, v);
        }   }
        else
        {   // assert(style == STYLE_CALM);
            if (psl & PSL_WC)
            {   strcat(mn, "ADDC     ");
            } else
            {   strcat(mn, "ADD      ");
            }
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",#");
            strcat(mn, v);
            if (operand == 1)
            {   sprintf(an, "%s++", calm);
            } else
            {   sprintf(an, "%s += $%s", calm, v);
        }   }
    acase 0x88: // @210
    case 0x89: // @211
    case 0x8A: // @212
    case 0x8B: // @213
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "ADDR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s += *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            if (psl & PSL_WC)
            {   strcat(mn, "ADDC     ");
            } else
            {   strcat(mn, "ADD      ");
            }
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s += *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x8C: // @214
    case 0x8D: // @215
    case 0x8E: // @216
    case 0x8F: // @217
        ABS_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "ADDA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s += *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            if (psl & PSL_WC)
            {   strcat(mn, "ADDC     ");
            } else
            {   strcat(mn, "ADD      ");
            }
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s += *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0x92: // @222
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "LPSU"); // or "LPSU    r0"
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     U,A");
        }
        strcpy(an, "PSU = (r0 & %01100111)");
    acase 0x93: // @223
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "LPSL"); // or "LPSL    r0"
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     L,A");
        }
        strcpy(an, "PSL = r0");
    acase 0x94: // @224
    case 0x95: // @225
    case 0x96: // @226
    case 0x97: // @227
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "DAR     r");
            strcat(mn, reg);
            sprintf(an, "r%s = BCD(r%s);", reg, reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "DA       ");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "%s = BCD(%s);", calm, calm);
        }
    acase 0x98: // @230
    case 0x99: // @231
    case 0x9A: // @232
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BCFR,");
            strcat(mn, ccstring[style][BRANCHCODE]);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "JUMP ");
            if (opcode == 0x98)
            {   strcat(mn, "NE ");
            } elif (opcode == 0x99)
            {   strcat(mn, "LE ");
            } else
            {   // assert(opcode == 0x9A);
                strcat(mn, "GE ");
        }   }
        strcat(mn, " ");
        strcat(mn, addressstring);
        sprintf(an, "%sgoto %s", ccfalse[BRANCHCODE], addressstring);
    acase 0x9B: // @233
        ZERO_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "ZBRR    "); // or "ZBRR,un "
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "JUMP     ");
        }
        strcat(mn, addressstring);
        sprintf(an, "goto %s", addressstring);
    acase 0x9C: // @234
    case 0x9D: // @235
    case 0x9E: // @236
        BRA_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BCFA,");
            strcat(mn, ccstring[style][BRANCHCODE]);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "JUMP ");
            if (opcode == 0x9C)
            {   strcat(mn, "NE ");
            } elif (opcode == 0x9D)
            {   strcat(mn, "LE ");
            } else
            {   // assert(opcode == 0x9E);
                strcat(mn, "GE ");
        }   }
        strcat(mn, " ");
        strcat(mn, addressstring);
        sprintf(an, "%sgoto %s", ccfalse[BRANCHCODE], addressstring);
    acase 0x9F: // @237
        BRA_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   if (psl & PSL_RS)
            {   strcat(mn, "BXA,r6  ");
                sprintf(an, "goto %s + r6", addressstring);
            } else
            {   strcat(mn, "BXA,r3  ");
                sprintf(an, "goto %s + r3", addressstring);
        }   }
        else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "JUMP     (D)+");
            if (psl & PSL_RS)
            {   sprintf(an, "goto %s + D'", addressstring);
            } else
            {   sprintf(an, "goto %s + D", addressstring);
        }   }
        strcat(mn, addressstring);
    acase 0xA0: // @240
    case 0xA1: // @241
    case 0xA2: // @242
    case 0xA3: // @243
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "SUBZ    r");
            strcat(mn, reg);
            sprintf(an, "r0 -= r%s", reg);
        } else
        {   // assert(style == STYLE_CALM);
            if (psl & PSL_WC)
            {   strcat(mn, "SUBC     A,");
            } else
            {   strcat(mn, "SUB      A,");
            }
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "A -= %s", calm);
        }
    acase 0xA4: // @244
    case 0xA5: // @245
    case 0xA6: // @246
    case 0xA7: // @247
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "SUBI,r");
            strcat(mn, reg);
            strcat(mn, " $");
            strcat(mn, v);
            if (operand == 1)
            {   sprintf(an, "r%s--", reg);
            } else
            {   sprintf(an, "r%s -= $%s", reg, v);
        }   }
        else
        {   // assert(style == STYLE_CALM);
            if (psl & PSL_WC)
            {   strcat(mn, "SUBB     ");
            } else
            {   strcat(mn, "SUB      ");
            }
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",#");
            strcat(mn, v);
            if (operand == 1)
            {   sprintf(an, "%s--", calm);
            } else
            {   sprintf(an, "%s -= $%s", calm, v);
        }   }
    acase 0xA8: // @250
    case 0xA9: // @251
    case 0xAA: // @252
    case 0xAB: // @253
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "SUBR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s -= *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            if (psl & PSL_WC)
            {   strcat(mn, "SUBB     ");
            } else
            {   strcat(mn, "SUB      ");
            }
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s -= *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0xAC: // @254
    case 0xAD: // @255
    case 0xAE: // @256
    case 0xAF: // @257
        ABS_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "SUBA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "r%s -= *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            if (psl & PSL_WC)
            {   strcat(mn, "SUBB     ");
            } else
            {   strcat(mn, "SUB      ");
            }
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "%s -= *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0xB0: // @260
    case 0xB1: // @261
    case 0xB2: // @262
    case 0xB3: // @263
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "WRTC    r");
            strcat(mn, reg);
            sprintf(an, "PORTC = r%s", reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     $CTRL,");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "PORTC = %s", calm);
        }
    acase 0xB4: // @264
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "TPSU    $");
            strcat(mn, v);
        } else
        {   // assert(style == STYLE_CALM);
            switch (operand)
            {
            case 0x20: // @40
                strcat(mn, "TEST     IOF");
            acase 0x40: // @100
                strcat(mn, "TEST     OUTPUT");
            acase 0x80: // @200
                strcat(mn, "TEST     INPUT");
            adefault:
                strcat(mn, "TEST     U,#");
                strcat(mn, v);
        }   }
        sprintf(an, "test bits $%s of PSU", v);
    acase 0xB5: // @265
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "TPSL    $");
            strcat(mn, v);
        } else
        {   // assert(style == STYLE_CALM);
            switch (operand)
            {
            case 0x01: // @1
                strcat(mn, "TEST     CARRY");
            acase 0x02: // @2
                strcat(mn, "TEST     LOGICOMP");
            acase 0x04: // @4
                strcat(mn, "TEST     OVERFLOW");
            acase 0x08: // @10
                strcat(mn, "TEST     WITHCARRY");
            acase 0x10: // @20
                strcat(mn, "TEST     BANK"); // BANK1 is acceptable as a synonym of BANK
            adefault:
                strcat(mn, "TEST     L,#");
                strcat(mn, v);
        }   }
        sprintf(an, "test bits $%s of PSL", v);
    acase 0xB8: // @270
    case 0xB9: // @271
    case 0xBA: // @272
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BSFR,");
            strcat(mn, ccstring[style][BRANCHCODE]);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "CALL ");
            if (opcode == 0xB8)
            {   strcat(mn, "NE ");
            } elif (opcode == 0xB9)
            {   strcat(mn, "LE ");
            } else
            {   // assert(opcode == 0xBA);
                strcat(mn, "GE ");
        }   }
        strcat(mn, " ");
        strcat(mn, addressstring);
        sprintf(an, "%sgosub %s", ccfalse[BRANCHCODE], addressstring);
    acase 0xBB: // @273
        ZERO_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "ZBSR    "); // or "ZBSR,un "
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "CALL     ");
        }
        strcat(mn, addressstring);
        sprintf(an, "gosub %s", addressstring);
    acase 0xBC: // @274
    case 0xBD: // @275
    case 0xBE: // @276
        BRA_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BSFA,");
            strcat(mn, ccstring[style][BRANCHCODE]);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "CALL ");
            if (opcode == 0xBC)
            {   strcat(mn, "NE ");
            } elif (opcode == 0xBD)
            {   strcat(mn, "LE ");
            } else
            {   // assert(opcode == 0xBE);
                strcat(mn, "GE ");
        }   }
        strcat(mn, " ");
        strcat(mn, addressstring);
        sprintf(an, "%sgosub %s", ccfalse[BRANCHCODE], addressstring);
    acase 0xBF: // @277
        BRA_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   if (psl & PSL_RS)
            {   strcat(mn, "BSXA,r6 ");
                sprintf(an, "gosub %s + r6", addressstring);
            } else
            {   strcat(mn, "BSXA,r3 ");
                sprintf(an, "gosub %s + r3", addressstring);
        }   }
        else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "CALL     (D)+");
            if (psl & PSL_RS)
            {   sprintf(an, "gosub %s + D'", addressstring);
            } else
            {   sprintf(an, "gosub %s + D", addressstring);
        }   }
        strcat(mn, addressstring);
    acase 0xC0: // @300
        strcat(mn, "NOP"); // same for both styles
        an[0] = 0;
    acase 0xC1: // @301
    case 0xC2: // @302
    case 0xC3: // @303
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "STRZ    r");
            strcat(mn, reg);
            sprintf(an, "r%s = r0", reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",A");
            sprintf(an, "%s = A", calm);
        }
    acase 0xC8: // @310
    case 0xC9: // @311
    case 0xCA: // @312
    case 0xCB: // @313
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "STRR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            strcat(mn, addressstring);
            sprintf(an, "*(%s) = r%s", addressstring, reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     ");
            strcat(mn, addressstring);
            strcat(mn, ",");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "*(%s) = %s", addressstring, calm);
        }
    acase 0xCC: // @314
    case 0xCD: // @315
    case 0xCE: // @316
    case 0xCF: // @317
        ABS_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "STRA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            strcat(mn, addressstring);
            sprintf(an, "*(%s) = r%s", addressstring, reg);
        } else
        {   // assert(style == STYLE_CALM)
            strcat(mn, "LOAD     ");
            strcat(mn, addressstring);
            strcat(mn, ",");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "*(%s) = %s", addressstring, calm);
        }
    acase 0xD0: // @320
    case 0xD1: // @321
    case 0xD2: // @322
    case 0xD3: // @323
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "RRL,r");
            strcat(mn, reg);
            sprintf(an, "r%s <<= 1", reg);
        } else
        {   // assert(style == STYLE_CALM);
            if (psl & PSL_WC)
            {   strcat(mn, "RLC      ");
            } else
            {   strcat(mn, "RL       ");
            }
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "%s <<= 1", calm);
        }
    acase 0xD4: // @324
    case 0xD5: // @325
    case 0xD6: // @326
    case 0xD7: // @327
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "WRTE,r");
            strcat(mn, reg);
            strcat(mn, " $");
            strcat(mn, v);
            sprintf(an, "IOPORT($%s) = r%s", v, reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     $");
            strcat(mn, v);
            strcat(mn, ",");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "IOPORT($%s) = %s", v, calm);
        }
    acase 0xD8: // @330
    case 0xD9: // @331
    case 0xDA: // @332
    case 0xDB: // @333
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BIRR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            strcat(mn, addressstring);
            if (operand == 0)
            {   sprintf(an, "r%s++", reg);
            } else
            {   sprintf(an, "if (++r%s != 0) goto %s", reg, addressstring);
        }   }
        else
        {   // assert(style == STYLE_CALM);
            CALMREG;
            if (operand == 0)
            {   strcat(mn, "INC      ");
                strcat(mn, calm);
                sprintf(an, "%s++", calm);
            } else
            {   strcat(mn, "INCJ,NE  ");
                strcat(mn, calm);
                strcat(mn, ",");
                strcat(mn, addressstring);
                sprintf(an, "if (++%s != 0) goto %s", calm, addressstring);
        }   }
    acase 0xDC: // @334
    case 0xDD: // @335
    case 0xDE: // @336
    case 0xDF: // @337
        BRA_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BIRA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "if (++r%s != 0) goto %s", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "INCJ,NE  ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "if (++%s != 0) goto %s", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0xE0: // @340
    case 0xE1: // @341
    case 0xE2: // @342
    case 0xE3: // @343
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "COMZ    r");
            strcat(mn, reg);
            sprintf(an, "compare r0 against r%s", reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "COMP     A,");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "compare A against %s", reg);
        }
    acase 0xE4: // @344
    case 0xE5: // @345
    case 0xE6: // @346
    case 0xE7: // @347
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "COMI,r");
            strcat(mn, reg);
            strcat(mn, " $");
            strcat(mn, v);
            sprintf(an, "compare r%s against $%s", reg, v);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "COMP     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",#");
            strcat(mn, v);
            sprintf(an, "compare %s against $%s", calm, v);
        }
    acase 0xE8: // @350
    case 0xE9: // @351
    case 0xEA: // @352
    case 0xEB: // @353
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "COMR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "compare r%s against *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "COMP     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "compare %s against *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0xEC: // @354
    case 0xED: // @355
    case 0xEE: // @356
    case 0xEF: // @357
        ABS_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "COMA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "compare r%s against *(%s)", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "COMP     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "compare %s against *(%s)", calm, addressstring);
        }
        strcat(mn, addressstring);
    acase 0xF0: // @360
    case 0xF1: // @361
    case 0xF2: // @362
    case 0xF3: // @363
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "WRTD    r");
            strcat(mn, reg);
            sprintf(an, "PORTD = r%s", reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "LOAD     $DATA,");
            CALMREG;
            strcat(mn, calm);
            sprintf(an, "PORTD = %s", calm);
        }
    acase 0xF4: // @364
    case 0xF5: // @365
    case 0xF6: // @366
    case 0xF7: // @367
        stcl_h(v, operand);
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "TMI,r");
            strcat(mn, reg);
            strcat(mn, "  $");
            strcat(mn, v);
            sprintf(an, "test bits $%s of r%s", v, reg);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "TEST     ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",#");
            strcat(mn, v);
            sprintf(an, "test bits $%s of %s", v, calm);
        }
    acase 0xF8: // @370
    case 0xF9: // @371
    case 0xFA: // @372
    case 0xFB: // @373
        REL_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BDRR,r");
            strcat(mn, reg);
            strcat(mn, " ");
            strcat(mn, addressstring);
            if (operand == 0)
            {   sprintf(an, "r%s--", reg);
            } else
            {   sprintf(an, "if (--r%s != 0) goto %s", reg, addressstring);
        }   }
        else
        {   // assert(style == STYLE_CALM);
            CALMREG;
            if (operand == 0)
            {   strcat(mn, "DEC      ");
                strcat(mn, calm);
                sprintf(an, "%s--", calm);
            } else
            {   strcat(mn, "DECJ,NE  ");
                strcat(mn, calm);
                strcat(mn, ",");
                strcat(mn, addressstring);
                sprintf(an, "if (--%s != 0) goto %s", calm, addressstring);
        }   }
    acase 0xFC: // @374
    case 0xFD: // @375
    case 0xFE: // @376
    case 0xFF: // @377
        BRA_EA_TRACE();
        if (style == STYLE_SIGNETICS)
        {   strcat(mn, "BDRA,r");
            strcat(mn, reg);
            strcat(mn, " ");
            sprintf(an, "if (--r%s != 0) goto %s", reg, addressstring);
        } else
        {   // assert(style == STYLE_CALM);
            strcat(mn, "DECJ,NE  ");
            CALMREG;
            strcat(mn, calm);
            strcat(mn, ",");
            sprintf(an, "if (--%s != 0) goto %s", calm, addressstring);
        }
        strcat(mn, addressstring);
    adefault: // $90, $91, $B6, $B7, $C4..$C7
        strcat(mn, "?");
        strcpy(an, "?");
}   }

EXPORT void cpu_setup(void)
{   mn[0]    =
    byte1[2] =
    byte2[2] =
    byte3[2] = EOS;
}

MODULE UBYTE cpuread(int address)
{   PERSIST UBYTE t;     // PERSISTent for speed
    PERSIST int   runit; // PERSISTent for speed

    /* assert(address <= 32767);

    Reads to memory from the UVI/PVI do not go through this routine.
    Strictly speaking, we should use this routine for ALL CPU read
    accesses, eg. reading opcodes/operands, relative addresses, etc.
    Currently, we only use it for absolute accesses: LODA, STRA, etc.

    Testing on an authentic Emerson Arcadia 2001 console has been
    undertaken to make the read/write property emulation as accurate as
    possible. However, this has not occurred for the Interton VC 4000.

    Reads of read-once registers, even if it is the second or later read,
    are not logged as illegal reads.

    Reads of unused ROM areas (eg. $0800..$0FFF for 2K ROMs) are not
    logged as illegal reads.

       1. Resolve read mirrors.
       2. Mark as read.
       3. Read the byte.
       4. If read-once
           set value
           log to user
       5. If EPROM
           log to user.
       6. If watchpoint
           log to user.
       7. If logged and pause on log
           pause */

    runit = FALSE;

    if (memflags[address] & MIRRORED_R)
    {   if (logreadwrites)
        {   getfriendly((int) mirror_r[address]);
            T1; textpen(TEXTPEN_LOG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_READMIRROR,
                    "Instruction at $%X (raster %d) read from mirrored address $%X (resolves to %s)!\n\n"
                ),
                (unsigned int) iar,
                (long int)     rast,
                (unsigned int) address,
                               friendly
            );
            T2;
            runit = TRUE;
        }
        address = (int) mirror_r[address];
    }

    memflags[address] |= COVERAGE_READ;

    if (machine == INSTRUCTOR && address == S_IOPORT && s_io == PARALLEL_MEMMAPPED)
    {   t = s_toggles;
    } elif (memflags[address] & RANDOM)
    {   /* assert(memmap == MEMMAP_F);
           assert(   address == E_RANDOM1E
                  || address == E_RANDOM1G
                  || address == E_RANDOM2); */
        if (recmode == RECMODE_PLAY)
        {   t = loadrand();
        } else
        {   if (connected == NET_SERVER)
            {   t = rand() % 256;
                DISCARD NetSendByte(t); // should check return code!
            } elif (connected == NET_CLIENT)
            {   t = NetReceiveByte(); // should check return code!
            } else
            {   t = rand() % 256;
            }
            if (recmode == RECMODE_RECORD)
            {   saverand(t);
    }   }   }
    elif (memflags[address] & NOREAD)
    /* It would be better to make a memread[32768] array, containing what
       values each read-only address returns. (Setting memory[] as
       appropriate wouldn't work, because of write-only registers.) */
    {   switch (machine)
        {
        case ARCADIA:
            t = 0xFF; // some really return eg. $40, $C0 instead of $FF
        acase INTERTON:
            // GRNDPRIX attempts to read from $1F6E..$1F70!
            // not sure whether NOISE ($1E80) is R/W or -/W! (assuming R/W)
            if (address == IE_NOISE)
            {   t = 0x0C; // assuming same behaviour as Elektor
            } elif
            (   // It would probably be better to handle these in setmemmap().
                (address >= 0x1F30 && address <= 0x1F3F)
             || (address >= 0x1F6E && address <= 0x1F7F)
             || (address >= 0x1FAE && address <= 0x1FBF) // this region is $00 for Elektor, so perhaps also for Interton?
            )
            {   t = 0xFF; // same as how MESS V2 initializes these regions
            } else
            {   t = 0x00;
                /* this is assumed rather than known behaviour, but:
                Interton BLACKJAC expects the high nybble (at least) of $1E8F to be $0! */
            }
        acase ELEKTOR:
            if (address >= 0x1000 && address <= 0x15FF)
            {   // assert(memmap == MEMMAP_E);
                t = 0xFF; // games which attempt to sense whether in mode "E"/"F" expect unavailable memory to return $FF
            } elif (address >= 0x1D80 && address <= 0x1DBF)
            {   t = 0x80;
                // TESTER2.PGM proves that $1D80 reads as $80.
            } elif (address >= 0x1DC0 && address <= 0x1DFF)
            {   t = 0xC0;
                // TESTER2.PGM proves that $1DC0 reads as $C0.
            } elif (address == IE_NOISE)
            {   t = 0x0C;
                // TESTER2.PGM proves that NOISE reads as $0C.
            } elif
            (   (address >= 0x1E8F && address <= 0x1E97)
             || (address >= 0x1E9C && address <= 0x1EA7)
             || (address >= 0x1EAF && address <= 0x1EB7)
             || (address >= 0x1EBC && address <= 0x1EC7)
             || (address >= 0x1ECF && address <= 0x1ED7)
             || (address >= 0x1EDC && address <= 0x1EE7)
             || (address >= 0x1EEF && address <= 0x1EF7)
             || (address >= 0x1EFC && address <= 0x1EFF)
            )
            {   t = 0x0F;
                // TESTER2.PGM proves that $1E93 reads as $0F.
            } else
            {   // eg. $1FAE..$1FBF, $1FC0..$1FC9
                // Elektor Labyrinth reads $1FB3..$1FB5 and expects them to be $00.
                // Also, TESTER2.PGM proves that $1FB8 reads as $00.
                // Elektor CarRace works best when SPRITES23CTRL reads as $00.
                t = 0;
            }
        adefault:
            t = 0; // not sure what unmapped PVI areas should return
        }

        if (logreadwrites)
        {   getfriendly(address);
            T1; textpen(TEXTPEN_LOG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_ILLEGALREAD,
                    "Instruction at $%X (raster %d) is reading $%02X from unreadable address %s!\n\n"
                ),
                (int) iar,
                rast,
                (int) t,
                friendly
            );
            T2;
            runit = TRUE;
    }   }
    elif
    (   (memflags[address] & BANKED)
     && !(psu & PSU_F)
    )
    {   // assert(machine == ASTROWARS || machine == GALAXIA);
        // assert(memflags[address] & VISIBLE);
        // assert(address >= 0x1800 && address <= 0x1BFF);
        t = g_bank[address - 0x1800];
    } else
    {   t = memory[address];
    }

    if (memflags[address] & READONCE)
    {   memory[address] = machines[machine].readonce;

        if (memflags[address] & WATCHPOINT)
        {   // we log this regardless of whether the value was actually changed,
            // and regardless of whether "watch reads?" is on or off.
            T1;
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_ENGINE_READONCE,
                    "Read-once register %s has been reset to $%X!\n" // yes, just one \n
                ),
                               friendly,
                (unsigned int) memory[address]
            );
            T2;
            runit = TRUE;
    }   }

    if (logeprom && (memflags[address] & EPROM))
    {   getfriendly(address);
        T1; textpen(TEXTPEN_LOG);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_READEPROM,
                "Instruction at $%X (raster %d) is reading $%02X from EPROM address %s!\n\n"
            ),
            (unsigned int) iar,
            (long int)     rast,
            (int)          t,
                           friendly
        );
        T2;
        runit = TRUE;
    }

    if
    (   (memflags[address] & WATCHPOINT)
     && watchreads
     && (   !(memflags[address] & BANKED)
         || (psu & PSU_F)
    )   )
    {   getfriendly(address);
        T1; textpen(TEXTPEN_LOG);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_ENGINE_HITWP,
                "Instruction at $%X (raster %d) is reading $%X from %s!\n" // yes, just one \n
            ),
            (unsigned int) iar,
            (long int)     rast,
                           t,
                           friendly
        );
        zprintf("\n");
        T2;
        runit = TRUE;
    }

    if (runit && pauselog)
    {   runto();
    }

    return t;
}

MODULE __inline void cpuwrite(int address, UBYTE data)
{   PERSIST int runit; // PERSISTent for speed

    /* assert(address < 32768);

    Writes to memory from the CPU always go through this routine.

    The game can never write to semi-mirrored registers, as they are
    all read-only. They can only be set by the PVI.

       1. Resolve write mirrors.
       2. Mark as written.
       3. If read-only
           log to user
           return
       4. If EPROM
           log to user
           return
       5. If watchpoint
           log to user
       6. If logged and pause on log
           pause
       7. Write the byte.

    handle mirroring */

    runit = FALSE;
    
    if (memflags[address] & MIRRORED_W)
    {   if (logreadwrites)
        {   getfriendly((int) mirror_w[address]);
            T1; textpen(TEXTPEN_LOG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_CPU_WRITEMIRROR,
                    "Instruction at $%X (raster %d) is writing $%X to mirrored address $%X (resolves to %s)!\n\n"
                ),
                (long unsigned int) iar,
                (long int)          rast,
                                    data,
                (long unsigned int) address,
                                    friendly
            );
            T2;
            runit = TRUE;
        }
        address = (int) mirror_w[address];
    }

    memflags[address] |= COVERAGE_WRITE;

    if
    (   (memflags[address] & NOWRITE)
     || (memflags[address] & EPROM)
    )
    {   if (logreadwrites && (memflags[address] & NOWRITE))
        {   getfriendly(address);
            T1; textpen(TEXTPEN_LOG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_CPU_ILLEGALWRITE,
                    "Instruction at $%X (raster %d) is attempting to write $%X to unwritable address %s!\n\n"
                ),
                (long unsigned int) iar,
                (long int)          rast,
                                    data,
                                    friendly
            );
            T2;
            runit = TRUE;
        }
        if (logeprom && (memflags[address] & EPROM))
        {   getfriendly(address);
            T1; textpen(TEXTPEN_LOG);
            zprintf
            (   GetCatalogStr(CatalogPtr,
                    MSG_CPU_WRITEEPROM,
                    "Instruction at $%X (raster %d) is attempting to write $%X to EPROM address %s!\n\n"
                ),
                (unsigned int) iar,
                (long int)     rast,
                               data,
                               friendly
            );
            T2;
            runit = TRUE;
        }

        if (runit && pauselog)
        {   runto();
        }
        return;
    }

    if
    (   (memflags[address] & WATCHPOINT)
     && (   !(memflags[address] & BANKED)
         || (psu & PSU_F)
        )
     && data != memory[address]
    )
    {   getfriendly(address);
        T1; textpen(TEXTPEN_LOG);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_CPU_HITWP,
                "Instruction at $%X (raster %d) is writing $%X to %s!\n\n"
            ),
            (long unsigned int) iar,
            (long int)          rast,
                                data,
                                friendly
        );
        T2;
        runit = TRUE;
    }

    if (runit && pauselog)
    {   runto();
    }

    if
    (   (memflags[address] & BANKED)
     && !(psu & PSU_F)
    )
    {   g_bank[address - 0x1800] = data;
    } elif (machine == INSTRUCTOR && address == S_IOPORT && s_io == PARALLEL_MEMMAPPED)
    {   s_glow = data;
    } else
    {   memory[address] = data;

        if (memflags[address] & FIELD)
        {   m_field[malzak_x / 16][(address - 0x1600) / 16] = data;
        }
        if (memflags[address] & VISIBLE)
        {   changed = TRUE;
        }
        if (memflags[address] & AUDIBLE)
        {   playsound(FALSE);
}   }   }

EXPORT void view_next(void)
{   getfriendly(iar);

    T1; textpen(TEXTPEN_TRACE);
    zprintf
    (   GetCatalogStr(CatalogPtr,
            MSG_ENGINE_NEXTINSTRUCTION,
            "Next instruction is %s at %s.\n\n"
        ),
        opcodename[memory[iar]],
        friendly
    );
    T2;
}

MODULE void do_runtoloopend(void)
{   getfriendly(iar);
    T1; textpen(TEXTPEN_DEBUG);
    zprintf
    (   GetCatalogStr(CatalogPtr,
            MSG_CPU_REACHEDLOOPEND,
            "Reached end of loop at %s.\n\n"
        ),
        friendly
    );
    T2;
    runto();
}
