/* pvi.c: Signetics 2636 PVI emulation module (host-independent)

This is how each frame is currently emulated:

VMAIN: now VRST is turned off
  0..  19 are  20 "margin" (above grid but visible) Y-pixels
  20..219 are 200 "grid"                            Y-pixels
 220..271 are  52 "margin" (below grid but visible) Y-pixels
now VRST is turned on
 272..291 are  20 "vertical retrace" (below grid)   Y-pixels
the true frame boundary is here
 292..311 are  20 "vertical retrace" (above grid)   Y-pixels

Almost but not quite the same as shown in the Interton VC 4000 Coding
 Guide. Doing it the same way as the Coding Guide doesn't seem to help.

Interton VC 4000: assumed behaviour is that the low nybbles of CONSOLE and
PnFOOKEYS are set to %1111, the same as is proven to occur on an Elektor
TVGC.

There is some special case code to deal with coin-ops. Eg.
Ami/WinGalaxia doesn't generate sprite-complete interrupts for Galaxia.
That is rather kludgy.

INCLUDES--------------------------------------------------------------- */

#ifdef AMIGA
    #include "amiga.h"
    #include <proto/locale.h> // GetCatalogStr()
#endif
#ifdef _lint
    #define WIN32
    #define _inline
#endif
#ifdef WIN32
    #include "ibm.h"
    #define EXEC_TYPES_H
#endif

#include "aa.h"
#define CATCOMP_NUMBERS
#define CATCOMP_BLOCK
#include "aa_strings.h"

#include <stdio.h>

// DEFINES----------------------------------------------------------------

// #define NOGRID
// whether to turn off grid

// #define NOHIDE
// whether you want to force sprites which are horizontally offscreen to be moved onsceen
// and force sprites which have no imagery (ie. all blank) to have imagery (ie. solid square)
// But note that it doesn't change the colour of sprites which are
// invisible because of their colour.

#define VMAIN 0

// EXPORTED VARIABLES-----------------------------------------------------

EXPORT       int   collx, colly;
EXPORT const UWORD pvi_spritedata[4] = { PVI_SPRITE0,
                                         PVI_SPRITE1,
                                         PVI_SPRITE2,
                                         PVI_SPRITE3 };
EXPORT       UBYTE bgcoll;
EXPORT       UWORD colltable[PVI_RASTLINES][PVI_XSIZE];
/* bits are:
  15..12: sprites #11..#8 respectively
  11.. 8: sprites  #7..#4 respectively
   7.. 4: sprites  #3..#0 respectively
       3: grid/playfield
   2.. 0: sprite foreground colour (wire OR'ed)

IMPORTED VARIABLES----------------------------------------------------- */

IMPORT FLAG   mmb,
              lmb;
IMPORT UBYTE  g_dips;
IMPORT ASCREEN screen[BOXWIDTH][BOXHEIGHT];
IMPORT ULONG  autofire,
              collisions,
              flagline,
              swapped;
IMPORT UBYTE* IOBuffer;
IMPORT FILE*  MacroHandle;
IMPORT TEXT   friendly[2 + 13 + 1 + 1];
IMPORT int    ax[2],
              ay[2],
              connected,
              console_start,
              console_a,
              console_b,
              console_reset,
              consoleopen,
              cpl,
              enhancestars,
              forceinterrupt,
              hostcontroller[2],
              inframe,
              key1,
              key2,
              key3,
              key4,
              latestsprite,
              p1bgcol[6],
              p2bgcol[6],
              p1sprcol[6],
              p2sprcol[6],
              paddleup,
              paddledown,
              paddleleft,
              paddleright,
              pvibase,
              machine,
              memmap,
              offset,
              offsetsprites,
              overcalc, // must be signed!
              rast,
              rastn,
              recmode,
              requirebutton,
              p1rumble,
              p2rumble,
              rumbling[2],
              shadowing,
              slice,
              spritesvisible,
              starfield,
              trace,
              usehacks,
              vblank,
              zinterrupt;
IMPORT UBYTE  psu,
              memory[32768],
              OutputBuffer[7],
              sx[2], sy[2];
IMPORT UWORD  console[4],
              jf[2],
              keypads[2][NUMKEYS],
              mirror_r[32768],
              mirror_w[32768];
IMPORT ULONG  downframes,
              frames,
              memflags[32768],
              totalframes;
IMPORT struct KeyTableStruct keytable[16];
IMPORT struct MachineStruct  machines[MACHINES];
IMPORT const int             from_m_spr[8],
                             from_g_w_spr[8],
                             from_g_x_spr[8];

#ifdef WIN32
    IMPORT int             CatalogPtr;
#endif
#ifdef AMIGA
    IMPORT LONG            longpens[PENS];
    IMPORT int             divided,
                        // remainder,
                           wide,
                           size;
    IMPORT FLAG            blockarexx;
    IMPORT UBYTE*          byteptr[MAXBOXHEIGHT];
    IMPORT UWORD*          wordptr[MAXBOXHEIGHT];
    IMPORT ULONG*          longptr[MAXBOXHEIGHT];
    IMPORT struct Catalog* CatalogPtr;
#endif

// MODULE VARIABLES-------------------------------------------------------

MODULE int    bgc,
              fgc,
              xx;
MODULE UBYTE  digitcolour,
              shady[PVI_XSIZE],
              t;

MODULE struct
{   int   x,
          majorrow,
          minorrow,
          size,
          starty;
    UBYTE colour,
          imagery;
#ifdef NOHIDE
    UBYTE previmagery;
#endif
    FLAG  dup;
} sprite[3][4];

// Each "character" is 12*20 pixels (3*5 "chunky 4*4" pixels).
MODULE const char led[20][12 + 1] =
{   "aaaabbbbcccc",
    "aaaabbbbcccc",
    "aaaabbbbcccc",
    "aaaabbbbcccc",
    "llll    dddd",
    "llll    dddd",
    "llll    dddd",
    "llll    dddd",
    "kkkkmmmmeeee",
    "kkkkmmmmeeee",
    "kkkkmmmmeeee",
    "kkkkmmmmeeee",
    "jjjj    ffff",
    "jjjj    ffff",
    "jjjj    ffff",
    "jjjj    ffff",
    "iiiihhhhgggg",
    "iiiihhhhgggg",
    "iiiihhhhgggg",
    "iiiihhhhgggg"
};

// MODULE FUNCTIONS-------------------------------------------------------

MODULE __inline void drawgrid(void);
MODULE void ie_drawdigit(int x, int y, int d, int line);
MODULE void ie_playerinput(int source, int dest);
MODULE __inline void ie_emuinput(void);
MODULE __inline void breakrastline(void);
MODULE __inline UBYTE pviread(int address);
MODULE void pvi_collisions(int rotations);
MODULE __inline void drawsprites(int whichpvi);
MODULE void newsprite(int whichpvi, int whichsprite);
MODULE void newmajorrow(int whichpvi, int whichsprite);
MODULE void newminorrow(int whichpvi, int whichsprite);
MODULE void drawminorrow(int whichpvi, int whichsprite);
MODULE void doshadows(void);
MODULE __inline void local_drawpixel(int x, int y, int colour);

// CODE-------------------------------------------------------------------

EXPORT void pvi(void)
{   PERSIST int i, x, y,
                whichsprite; // all PERSISTent for speed

    inframe = TRUE;
    if (shadowing)
    {   for (x = 0; x < PVI_XSIZE; x++)
        {   shady[x] = 16;
    }   }

    // "The data is valid only at vertical reset." - 2636 PVI datasheet, p. 10.
    // But enabling these lines breaks Interton BOXING :-(
    // pviwrite(PVI_P1PADDLE, 0, TRUE);
    // pviwrite(PVI_P2PADDLE, 0, TRUE);

    switch (machine)
    {
    case INTERTON:
    case ELEKTOR:
        ie_emuinput();
    acase ASTROWARS:
        g_emuinput();
        astrowars_drawscreen();
    acase GALAXIA:
        g_emuinput();
        galaxia_drawscreen();
    acase MALZAK:
        if (usehacks && memory[0x100A] == 0x06)
        {   offsetsprites = 68;
        } else
        {   offsetsprites =  0;
        }
        m_emuinput();
        m_drawscreen();
    }

    for (whichsprite = 0; whichsprite < 4; whichsprite++)
    {   sprite[0][whichsprite].starty           = ((UBYTE) (pviread(pvi_spritedata[whichsprite] + 12) + 1)); // SPRITEnAY
        sprite[0][whichsprite].dup              = FALSE;
        sprite[0][whichsprite].majorrow         = -1;
        if (machine == MALZAK || machine == GALAXIA)
        {   pvibase += 0x100; // switch to 2nd PVI
            sprite[1][whichsprite].starty       = ((UBYTE) (pviread(pvi_spritedata[whichsprite] + 12) + 1)); // SPRITEnAY
            sprite[1][whichsprite].dup          = FALSE;
            sprite[1][whichsprite].majorrow     = -1;
            if (machine == GALAXIA)
            {   pvibase += 0x100; // switch to 3rd PVI
                sprite[2][whichsprite].starty   = ((UBYTE) (pviread(pvi_spritedata[whichsprite] + 12) + 1)); // SPRITEnAY
                sprite[2][whichsprite].dup      = FALSE;
                sprite[2][whichsprite].majorrow = -1;
                pvibase = 0x1500; // switch to 1st PVI
            } else
            {   // assert(machine == MALZAK);
                pvibase = 0x1400; // switch to 1st PVI
    }   }   }

    if (machine == ASTROWARS || machine == GALAXIA)
    {   for (i = 0; i <= 2; i++)
        {   for (whichsprite = 0; whichsprite < 4; whichsprite++)
            {    sprite[i][whichsprite].starty += 2;
    }   }   }

    for (rast = 0; rast < PVI_RASTLINES; rast++)
    {   breakrastline();

        slice = cpl - overcalc;
        cpu();
        CHECKINPUT;

        // blank the current rastline
        // only doing this for rast < IE_BOXHEIGHT isn't good enough;
        // it needs to be done for every rastline
        if (machine != MALZAK)
        {   for (x = 0; x < PVI_XSIZE; x++)
            {   colltable[rast][x] = 0;
        }   }

        if (rast == VMAIN)
        {   if (machine == MALZAK)
            {   bgcoll = (memory[0x14CA] & 0xF0) | ((memory[0x15CA] & 0xF0) >> 4); // are malzak collisions late by a frame?
            }
            pviwrite(PVI_SPRITECOLLIDE , 0, TRUE); // clear VRST and sprite-sprite collision bits
            pviwrite(PVI_SPRITECOLLIDE2, 0, TRUE); // clear VRST and sprite-sprite collision bits
            pviwrite(PVI_SPRITECOLLIDE3, 0, TRUE); // clear VRST and sprite-sprite collision bits
            pviwrite(PVI_SPRITECOLLIDE4, 0, TRUE); // clear VRST and sprite-sprite collision bits
            pviwrite(PVI_BGCOLLIDE     , 0, TRUE); // clear sprite complete and sprite-bg collision bits
            pviwrite(PVI_BGCOLLIDE2    , 0, TRUE); // clear sprite complete and sprite-bg collision bits
            pviwrite(PVI_BGCOLLIDE3    , 0, TRUE); // clear sprite complete and sprite-bg collision bits
            pviwrite(PVI_BGCOLLIDE4    , 0, TRUE); // clear sprite complete and sprite-bg collision bits
            if (machine == MALZAK || machine == GALAXIA)
            {   pvibase += 0x100; // switch to 2nd PVI
                pviwrite(PVI_SPRITECOLLIDE , 0, TRUE); // clear VRST and sprite-sprite collision bits
                pviwrite(PVI_SPRITECOLLIDE2, 0, TRUE); // clear VRST and sprite-sprite collision bits
                pviwrite(PVI_SPRITECOLLIDE3, 0, TRUE); // clear VRST and sprite-sprite collision bits
                pviwrite(PVI_SPRITECOLLIDE4, 0, TRUE); // clear VRST and sprite-sprite collision bits
                pviwrite(PVI_BGCOLLIDE     , 0, TRUE); // clear sprite complete and sprite-bg collision bits
                pviwrite(PVI_BGCOLLIDE2    , 0, TRUE); // clear sprite complete and sprite-bg collision bits
                pviwrite(PVI_BGCOLLIDE3    , 0, TRUE); // clear sprite complete and sprite-bg collision bits
                pviwrite(PVI_BGCOLLIDE4    , 0, TRUE); // clear sprite complete and sprite-bg collision bits
                if (machine == GALAXIA)
                {   pvibase += 0x100; // switch to 3rd PVI
                    pviwrite(PVI_SPRITECOLLIDE , 0, TRUE); // clear VRST and sprite-sprite collision bits
                    pviwrite(PVI_SPRITECOLLIDE2, 0, TRUE); // clear VRST and sprite-sprite collision bits
                    pviwrite(PVI_SPRITECOLLIDE3, 0, TRUE); // clear VRST and sprite-sprite collision bits
                    pviwrite(PVI_SPRITECOLLIDE4, 0, TRUE); // clear VRST and sprite-sprite collision bits
                    pviwrite(PVI_BGCOLLIDE     , 0, TRUE); // clear sprite complete and sprite-bg collision bits
                    pviwrite(PVI_BGCOLLIDE2    , 0, TRUE); // clear sprite complete and sprite-bg collision bits
                    pviwrite(PVI_BGCOLLIDE3    , 0, TRUE); // clear sprite complete and sprite-bg collision bits
                    pviwrite(PVI_BGCOLLIDE4    , 0, TRUE); // clear sprite complete and sprite-bg collision bits
                    pvibase = 0x1500; // switch to 1st PVI
                } else
                {   // assert(machine == MALZAK);
                    pvibase = 0x1400; // switch to 1st PVI
            }   }
            psu &= ~(PSU_S);
        }

        if (rast < machines[machine].height && (machine == INTERTON || machine == ELEKTOR))
        {   drawgrid();

            if ((rast >= 20 && rast <=  39) || (rast >= 200 && rast <= 219))
            {   if (pviread(PVI_SCORECTRL) & 1)
                {   y = 200;
                } else
                {   y =  20;
                }
                if (rast == y)
                {   if (pviread(PVI_BGCOLOUR) & 8)
                    {   digitcolour = (pviread(PVI_BGCOLOUR) >> 4) & 7;
                    } else
                    {   digitcolour = 0; // white
                }   }
                // this is cached for eg. LASERATT game over screen
                if (rast >= y && rast < y + 20) // 20 rastlines is the height of each digit
                {   ie_drawdigit(    32 - HIDDEN_X + 28, y, pviread(PVI_SCORELT) >>   4, rast - y);
                    ie_drawdigit(    32 - HIDDEN_X + 44, y, pviread(PVI_SCORELT) &  0xf, rast - y);
                    if (pviread(PVI_SCORECTRL) & 2)
                    {   ie_drawdigit(32 - HIDDEN_X + 60, y, pviread(PVI_SCORERT) >>   4, rast - y);
                        ie_drawdigit(32 - HIDDEN_X + 76, y, pviread(PVI_SCORERT) &  0xf, rast - y);
                    } else
                    {   ie_drawdigit(32 - HIDDEN_X + 76, y, pviread(PVI_SCORERT) >>   4, rast - y);
                        ie_drawdigit(32 - HIDDEN_X + 92, y, pviread(PVI_SCORERT) &  0xf, rast - y);
        }   }   }   }

        // sprites raster must be drawn AFTER the background raster

        drawsprites(0);
        if (machine == MALZAK || machine == GALAXIA)
        {   pvibase += 0x100; // switch to 2nd PVI
            drawsprites(1);
            if (memmap == MEMMAP_GALAXIA)
            {   pvibase += 0x100; // switch to 3rd PVI
                drawsprites(2);
                pvibase = 0x1500;
            } else
            {   // assert(machine == MALZAK);
                pvibase = 0x1400;
        }   }

        // Once a collision is registered, does it remain set
        // for the remainder of the frame, or only the current
        // rastline? (We're assuming it's for the remainder of
        // the frame.)

        if (collisions && machine != ASTROWARS)
        {   pvi_collisions(0);
            if (machine == MALZAK || machine == GALAXIA)
            {   pvibase += 0x100; // switch to 2nd PVI
                pvi_collisions(4);
                if (machine == GALAXIA)
                {   pvibase += 0x100; // switch to 3rd PVI
                    pvi_collisions(8);
                    pvibase = 0x1500; // switch to 1st PVI
                } else
                {   // assert(machine == MALZAK);
                    pvibase = 0x1400; // switch to 1st PVI
        }   }   }

        if (rast == vblank)
        {   // this point is the start of vertical blank

            // ie_emuinput(); was here from V7.11-7.41, but it breaks
            // SHOOTGAL/HUNTING.

            pviwrite(PVI_SPRITECOLLIDE , (UBYTE) (pviread(PVI_SPRITECOLLIDE ) | 0x40), TRUE); // set VRST bit
            pviwrite(PVI_SPRITECOLLIDE2, (UBYTE) (pviread(PVI_SPRITECOLLIDE2) | 0x40), TRUE); // set VRST bit
            pviwrite(PVI_SPRITECOLLIDE3, (UBYTE) (pviread(PVI_SPRITECOLLIDE3) | 0x40), TRUE); // set VRST bit
            pviwrite(PVI_SPRITECOLLIDE4, (UBYTE) (pviread(PVI_SPRITECOLLIDE4) | 0x40), TRUE); // set VRST bit

            if (machine == MALZAK || machine == GALAXIA)
            {   pvibase += 0x100; // 2nd PVI
                pviwrite(PVI_SPRITECOLLIDE , (UBYTE) (pviread(PVI_SPRITECOLLIDE ) | 0x40), TRUE); // set VRST bit
                pviwrite(PVI_SPRITECOLLIDE2, (UBYTE) (pviread(PVI_SPRITECOLLIDE2) | 0x40), TRUE); // set VRST bit
                pviwrite(PVI_SPRITECOLLIDE3, (UBYTE) (pviread(PVI_SPRITECOLLIDE3) | 0x40), TRUE); // set VRST bit
                pviwrite(PVI_SPRITECOLLIDE4, (UBYTE) (pviread(PVI_SPRITECOLLIDE4) | 0x40), TRUE); // set VRST bit
                if (machine == GALAXIA)
                {   pvibase += 0x100; // 3rd PVI
                    pviwrite(PVI_SPRITECOLLIDE , (UBYTE) (pviread(PVI_SPRITECOLLIDE ) | 0x40), TRUE); // set VRST bit
                    pviwrite(PVI_SPRITECOLLIDE2, (UBYTE) (pviread(PVI_SPRITECOLLIDE2) | 0x40), TRUE); // set VRST bit
                    pviwrite(PVI_SPRITECOLLIDE3, (UBYTE) (pviread(PVI_SPRITECOLLIDE3) | 0x40), TRUE); // set VRST bit
                    pviwrite(PVI_SPRITECOLLIDE4, (UBYTE) (pviread(PVI_SPRITECOLLIDE4) | 0x40), TRUE); // set VRST bit
                    pvibase = 0x1500; // 1st PVI
                    psu |= PSU_S & g_dips;
                } else
                {   // assert(machine == MALZAK);
                    pvibase = 0x1400; // 1st PVI
                    psu |= PSU_S;
            }   }
            elif (machine == ASTROWARS)
            {   psu |= PSU_S & g_dips;
            } else
            {   psu |= PSU_S;
            }

            if (!(psu & PSU_II) || forceinterrupt) // forceinterrupt of 1 or 2
            {   zinterrupt = TRUE;

#ifdef DEBUGINTERRUPTS
                T1; textpen(TEXTPEN_VERBOSE);
                zprintf("Generated vblank interrupt at rastline %d.\n", rast);
                T2;
            } else
            {   T1; textpen(TEXTPEN_VERBOSE);
                zprintf("Suppressed vblank interrupt at rastline %d.\n", rast);
                T2;
#endif
    }   }   }

#ifdef SHOWRUMBLE
    if (rumbling[0])
    {   for (y = 0; y < machines[machine].height; y++)
        {   local_drawpixel(0, y, 15 - screen[0][y]);
    }   }
    if (rumbling[1])
    {   for (y = 0; y < machines[machine].height; y++)
        {   local_drawpixel(machines[machine].width - 1, y, 15 - screen[machines[machine].width - 1][y]);
    }   }
#endif

    endofframe();
}

MODULE __inline void breakrastline(void)
// this is exactly the same as the uvi.c version, we could combine them
// (at some cost in speed)
{   if (rast == rastn)
    {   T1; textpen(TEXTPEN_DEBUG);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_XVI_REACHEDRASTLINEN,
                "Reached raster line %ld.\n\n"
            ),
            (long int) rast
        );
        T2;
        runto();
}   }

MODULE __inline void drawgrid(void)
{   PERSIST int gridline[220] =                                                         // Interton/Elektor screen is vertically as follows:
    {   20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, 20, //   0.. 19 top (visible) margin
         0,  0,                                                                         //  20.. 21 thin
                 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1, //  22.. 39 thick
         2,  2,                                                                         //  40.. 41 thin
                 3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3, //  42.. 59 thick
         4,  4,                                                                         //  60.. 61 thin
                 5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5, //  62.. 79 thick
         6,  6,                                                                         //  80.. 81 thin
                 7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7, //  82.. 99 thick
         8,  8,                                                                         // 100..101 thin
                 9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9, // 102..119 thick
        10, 10,                                                                         // 120..121 thin
                11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, // 122..139 thick
        12, 12,                                                                         // 140..141 thin
                13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, // 142..159 thick
        14, 14,                                                                         // 160..161 thin
                15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, 15, // 162..179 thick
        16, 16,                                                                         // 180..181 thin
                17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, 17, // 182..199 thick
        18, 18,                                                                         // 200..201 thin
                19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19  // 202..219 thick
                                                                                        // 220..271 bottom visible margin
                                                                                        // 272..311 bottom invisible margin
    };
    PERSIST int i, j, k,
                m,
                w,
                x; // PERSISTent for speed

    // assert(rast < IE_BOXHEIGHT);

    // it would be better not to call pviread() twice
    //  (will trigger watchpoints too often)
    if (flagline)
    {   if (pviread(PVI_BGCOLOUR) & 8) // background/grid enable
        {   fgc = 15 - ((pviread(PVI_BGCOLOUR) & 0x70) >> 4);
            bgc = 15 - ( pviread(PVI_BGCOLOUR) &    7);
            the_drawline(bgc, rast);
        } else
        {   bgc = 15; // dark black (black)
            the_drawline(bgc, rast);
            doshadows();
            return;
    }   }
    else
    {   if (pviread(PVI_BGCOLOUR) & 8) // background/grid enable
        {   fgc =  7 - ((pviread(PVI_BGCOLOUR) & 0x70) >> 4);
            bgc =  7 -  (pviread(PVI_BGCOLOUR) &    7);
            the_drawline(bgc, rast);
        } else
        {   bgc =  7; // black
            the_drawline(bgc, rast);
            doshadows();
            return;
    }   }

    doshadows();

#ifndef NOGRID
    if (rast < 20 || rast > 219)
#endif
        return;

    if
    (   (machine == INTERTON || memmap == MEMMAP_F)
     && (memory[IE_NOISE] & 0x20)
    )
    {   if (fgc >= 8)
        {   fgc = 15 - fgc; // 0..7
        } else
        {   fgc =  7 - fgc; // 0..7
    }   }

    i = gridline[rast];                 //  i       is     0..19
    k = pviread(PVI_HORIZ1 + (i >> 2)); // (i >> 2) yields 0.. 4
    switch (k & 0xC0)
    {
    case  0x40:
        w = 2;
    acase 0xC0:
        w = 4;
        adefault: // $00, $80
        w = 1;
    break;
    }

    switch (i & 3)                   // (i &  3) yields 0.. 3
    {
    case 0:
        if (k & 1)
        {   w = 8;
        }
    acase 1:
        if (((rast - 20) % 40) <= 10)
        {   if (k & 2)
            {   w = 8;
        }   }
        else
        {   if (k & 4)
            {   w = 8;
        }   }
    acase 2:
        if (k & 8)
        {   w = 8;
        }
    acase 3:
        if (((rast - 20) % 40) <= 30)
        {   if (k & 16)
            {   w = 8;
        }   }
        else
        {   if (k & 32)
            {   w = 8;
        }   }
    break;
    }

    // j goes   0,  1,  2,  3,  4,  5,  6,  7,   8,   9,  10,  11,  12,  13,  14,  15
    // m goes 128, 64, 32, 16,  8,  4,  2,  1, 128,  64,  32,  16,   8,   4,   2,   1
    // x goes  32, 42, 50, 58, 66, 74, 82, 90,  98, 106, 114, 122, 130, 138, 146, 154
    for
    (   x = 32, j = 0, m = 128; // was x = 31 until V4.62
        j < 16;
        j++, x += 8, m >>= 1
    )
    {   if (pviread(PVI_VERTGRID + (i * 2) + (j >> 3)) & m)
        {   for (xx = x; xx < x + w; xx++)
            {   colltable[rast][xx] |= 8;
                if (screen[xx - HIDDEN_X][rast] != (UBYTE) fgc)
                {   local_drawpixel
                    (   xx - HIDDEN_X,
                        rast,
                        fgc
                    );
                    if (shadowing)
                    {   shady[xx - HIDDEN_X + 1] = 15 - fgc;
        }   }   }   }
        if (j == 7)
        {   m = 256;
}   }   }

/* Gridlines that are the same colour as the background (ie. invisible)
   don't register collisions. This seems to be the correct behaviour.

     -0-  -1-  -2-  -3-  -4-  -5-  -6-  -7-  -8-  -9-  -A-  -B-  -C-  -D-  -E-  -F-
abc  ###  ..#  ###  ###  #.#  ###  ###  ###  ###  ###  ###  #..  ###  ..#  ###  ###
l d  # #  . #  ..#  . #  # #  # .  # .  . #  # #  # #  # #  # .  # .  . #  # .  # .
kme  #.#  ..#  ###  ###  ###  ###  ###  ..#  ###  ###  ###  ###  #..  ###  ###  ###
j f  # #  . #  #..  . #  . #  . #  # #  . #  # #  . #  # #  # #  # .  # #  # .  # .
ihg  ###  ..#  ###  ###  ..#  ###  ###  ..#  ###  ###  #.#  ###  ###  ###  ###  #..

The Interton/Elektor handles $A..$F values by drawing nothing.
The capital $B looks the same as 8
and capital $D looks the same as 0
so we would use lowercase "b" and "d". */

MODULE void ie_drawdigit(int x, int y, int d, int line)
{   // bit $1000: 'm' segment
    // bit $0800: 'l' segment
    // bit $0400: 'k' segment
    // bit $0200: 'j' segment
    // bit $0100: 'i' segment
    // bit $0080: 'h' segment
    // bit $0040: 'g' segment
    // bit $0020: 'f' segment
    // bit $0010: 'e' segment
    // bit $0008: 'd' segment
    // bit $0004: 'c' segment
    // bit $0002: 'b' segment
    // bit $0001: 'a' segment
    PERSIST const UWORD digit_to_segment[10] = // [16] for hex support
    {   0x0fff, // 0
        0x007c, // 1
        0x17df, // 2
        0x15ff, // 3
        0x1c7d, // 4
        0x1df7, // 5
        0x1ff7, // 6
        0x007f, // 7
        0x1fff, // 8
        0x1dff  // 9
/*    , 0x1f7f, // A
        0x1ff1, // B
        0x0fc7, // C
        0x17fc, // D
        0x1fd7, // E
        0x1f17  // F
*/
    };
    int j;

    // assert(d >=  0);
    // assert(d <= 15);
    if (d >= 10)
    {   return;
    }

    for (j = 0; j < 12; j++)
    {   if
        (   digit_to_segment[d] & (1 << (led[line][j] - 'a'))
        )
        {   // colltable[rast][x + j + HIDDEN_X] |= 8;
            // Based on Catapult, Destroyer and Submarine+Racing (all
            //  for Elektor), the real PVI probably doesn't involve the
            //  digits in collision detection. So we have changed our
            //  behaviour accordingly as of V12.62.
            if (screen[x + j][y + line] != digitcolour)
            {   local_drawpixel(x + j, y + line, digitcolour);
}   }   }   }

/* "Unlike SPRITEnAX and SPRITEnAY, SPRITEnBX and SPRITEnBY may
be changed duiring the scan [(frame)] and such changes will be effected
on the current scan [(frame)]. SPRITEnBX is sampled during each HRST
[(horizontal reset)]. Thus, if SPRITEnBX is changed during the
display of a duplicate, a portion of the object will be
displaced horizontally. For proper operation, SPRITEnBX should
be changed only *after* the 'Object Video Completion' status
bit [in BGCOLLIDE] indicates completion of the object.
SPRITEnBY is sampled just prior to displaying the last line of
the object. To effect a change in the vertical offset to the
next duplicate, SPRITEnBY must be changed *before* the 'Object
Video Completion' status bit [in BGCOLLIDE] indicates
completion of the object." - PVI manual, p. 5. */

EXPORT void pviwrite(signed int address, UBYTE data, FLAG ispvi)
{   /* Writes to memory from the PVI always go through this
       routine (so that watchpoints work properly). */

    if (ispvi)
    {   address += pvibase;
    }

    if ((memflags[address] & WATCHPOINT) && data != memory[address])
    {   DISCARD getfriendly(address);
        T1; textpen(TEXTPEN_DEBUG);
        zprintf
        (   GetCatalogStr(CatalogPtr,
                MSG_PVI_HITWP,
                "PVI is writing $%X to %s at raster %ld!\n\n"
            ),
                       data,
            (char *)   friendly,
            (long int) rast
        );
        T2;
        runto();
    }

    memory[address] = data;
}

// We could do uviread() too.
MODULE __inline UBYTE pviread(int address)
{   return memory[address + pvibase];
}

MODULE void ie_playerinput(int source, int dest)
{   PERSIST UWORD jg; // PERSISTent for speed

    // dest is which side   (0 or 1) you want to set the registers of.
    // source is which side (0 or 1) you want to use to do it.

    if (recmode == RECMODE_PLAY)
    {   // I_P1PADDLE is $1FCC, I_P2PADDLE is $1FCD
        t = IOBuffer[offset++];
        pviwrite(PVI_P1PADDLE + dest, t, TRUE);
        if (psu & PSU_F) // paddle interpolation bit
        {   ay[dest] = sy[dest] = t;
        } else
        {   ax[dest] = sx[dest] = t;
        }
        t = IOBuffer[offset++];
        pviwrite(IE_P1LEFTKEYS   + (dest * 4), (UBYTE) ( (t & 0xF0)       | 0x0F), FALSE); // bits 7..4 -> bits  7..4 (masking is indeed necessary)
        pviwrite(IE_P1MIDDLEKEYS + (dest * 4), (UBYTE) (((t & 0x0F) << 4) | 0x0F), FALSE); // bits 3..0 -> bits  7..4
        t = IOBuffer[offset++];
        pviwrite(IE_P1RIGHTKEYS  + (dest * 4),            t                      , FALSE);
    } else
    {   // assert(recmode == RECMODE_NORMAL || recmode == RECMODE_RECORD);

        if
        (   (dest == 0 && connected == NET_CLIENT)
         || (dest == 1 && connected == NET_SERVER)
        )
        {   return;
        }

        if
        (   hostcontroller[source] == CONTROLLER_1STDJOY
         || hostcontroller[source] == CONTROLLER_1STDPAD
      // || hostcontroller[source] == CONTROLLER_1STAJOY
         || hostcontroller[source] == CONTROLLER_1STAPAD
        )
        {   jg = jf[0];
        } elif
        (   hostcontroller[source] == CONTROLLER_2NDDJOY
         || hostcontroller[source] == CONTROLLER_2NDDPAD
         || hostcontroller[source] == CONTROLLER_2NDAJOY
         || hostcontroller[source] == CONTROLLER_2NDAPAD
        )
        {   jg = jf[1];
        } else
        {   jg = 0;
        }

        memory[IE_P1LEFTKEYS   + (dest * 4)] =
        memory[IE_P1MIDDLEKEYS + (dest * 4)] =
        memory[IE_P1RIGHTKEYS  + (dest * 4)] = 0x0F;
        if (autofire)
        {   if
            (   (   !requirebutton
                 || (jg & JOYFIRE1)
                 || KeyDown(keypads[source][0])
                 || (   hostcontroller[source] == CONTROLLER_TRACKBALL
                     && lmb
                )   )
             && ((frames % totalframes) < downframes)
            )
            {   // 1st firebutton
                memory[keytable[key1].ie_address + (dest * 4)] |= keytable[key1].ie_mask;
        }   }
        else
        {   if
            (   (jg & JOYFIRE1)
             || KeyDown(keypads[source][0])
             || (hostcontroller[source] == CONTROLLER_TRACKBALL && lmb)
            )
            {   // 1st firebutton
                memory[keytable[key1].ie_address + (dest * 4)] |= keytable[key1].ie_mask;
        }   }
        if
        (   (jg & JOYFIRE2)
         || (KeyDown(keypads[source][21]))
         || (hostcontroller[source] == CONTROLLER_TRACKBALL && mmb)
        )
        {   // 2nd firebutton
            memory[keytable[key2].ie_address + (dest * 4)] |= keytable[key2].ie_mask;
        }
        if
        (   (jg & JOYFIRE3)
         || (KeyDown(keypads[source][22]))
        )
        {   // 3rd firebutton
            memory[keytable[key3].ie_address + (dest * 4)] |= keytable[key3].ie_mask;
        }
        if
        (   (jg & JOYFIRE4)
         || (KeyDown(keypads[source][23]))
        )
        {   // 4th firebutton
            memory[keytable[key4].ie_address + (dest * 4)] |= keytable[key4].ie_mask;
        }

        // left column
        t =  memory[IE_P1LEFTKEYS + (dest * 4)];
        t |= (UBYTE) (0x80 * KeyDown(keypads[source][ 1])); // "1"  key (Interton) or "RCAS"  key (Elektor)
        t |= (UBYTE) (0x40 * KeyDown(keypads[source][ 4])); // "4"  key (Interton) or "BP1/2" key (Elektor)
        t |= (UBYTE) (0x20 * KeyDown(keypads[source][ 7])); // "7"  key (Interton) or "PC"    key (Elektor)
        t |= (UBYTE) (0x10 * KeyDown(keypads[source][10])); // "Cl" key (Interton) or "-"     key (Elektor)
        if (ay[dest] <=  64)
        {   if   (paddleup    ==  1) t |= 0x80;
            elif (paddleup    ==  4) t |= 0x40;
            elif (paddleup    ==  7) t |= 0x20;
            elif (paddleup    == 10) t |= 0x10;
        } elif (ay[dest] >= 192)
        {   if   (paddledown  ==  1) t |= 0x80;
            elif (paddledown  ==  4) t |= 0x40;
            elif (paddledown  ==  7) t |= 0x20;
            elif (paddledown  == 10) t |= 0x10;
        }
        if (ax[dest] <=  64)
        {   if   (paddleleft  ==  1) t |= 0x80;
            elif (paddleleft  ==  4) t |= 0x40;
            elif (paddleleft  ==  7) t |= 0x20;
            elif (paddleleft  == 10) t |= 0x10;
        } elif (ax[dest] >= 192)
        {   if   (paddleright ==  1) t |= 0x80;
            elif (paddleright ==  4) t |= 0x40;
            elif (paddleright ==  7) t |= 0x20;
            elif (paddleright == 10) t |= 0x10;
        }
        pviwrite(IE_P1LEFTKEYS    + (dest * 4), t, FALSE);

        // middle column
        t =  memory[IE_P1MIDDLEKEYS + (dest * 4)];
        t |= (UBYTE) (0x80 * KeyDown(keypads[source][ 2])); // "2"  key (Interton) or "WCAS" key (Elektor)
        t |= (UBYTE) (0x40 * KeyDown(keypads[source][ 5])); // "5"  key (Interton) or "REG"  key (Elektor)
        t |= (UBYTE) (0x20 * KeyDown(keypads[source][ 8])); // "8"  key (Interton) or "MEM"  key (Elektor)
        t |= (UBYTE) (0x10 * KeyDown(keypads[source][11])); // "0"  key (Interton) or "+"    key (Elektor)
        if (ay[dest] <=  64)
        {   if   (paddleup    ==  2) t |= 0x80;
            elif (paddleup    ==  5) t |= 0x40;
            elif (paddleup    ==  8) t |= 0x20;
            elif (paddleup    ==  0) t |= 0x10;
        } elif (ay[dest] >= 192)
        {   if   (paddledown  ==  2) t |= 0x80;
            elif (paddledown  ==  5) t |= 0x40;
            elif (paddledown  ==  8) t |= 0x20;
            elif (paddledown  ==  0) t |= 0x10;
        }
        if (ax[dest] <=  64)
        {   if   (paddleleft  ==  2) t |= 0x80;
            elif (paddleleft  ==  5) t |= 0x40;
            elif (paddleleft  ==  8) t |= 0x20;
            elif (paddleleft  ==  0) t |= 0x10;
        } elif (ax[dest] >= 192)
        {   if   (paddleright ==  2) t |= 0x80;
            elif (paddleright ==  5) t |= 0x40;
            elif (paddleright ==  8) t |= 0x20;
            elif (paddleright ==  0) t |= 0x10;
        }
        pviwrite(IE_P1MIDDLEKEYS  + (dest * 4), t, FALSE);

        // right column
        t =  memory[IE_P1RIGHTKEYS + (dest * 4)];
        t |= (UBYTE) (0x80 * KeyDown(keypads[source][ 3])); // "3"  key (Interton) or "C" key (Elektor)
        t |= (UBYTE) (0x40 * KeyDown(keypads[source][ 6])); // "6"  key (Interton) or "8" key (Elektor)
        t |= (UBYTE) (0x20 * KeyDown(keypads[source][ 9])); // "9"  key (Interton) or "4" key (Elektor)
        t |= (UBYTE) (0x10 * KeyDown(keypads[source][12])); // "En" key (Interton) or "0" key (Elektor)
        if (ay[dest] <=  64)
        {   if   (paddleup    ==  3) t |= 0x80;
            elif (paddleup    ==  6) t |= 0x40;
            elif (paddleup    ==  9) t |= 0x20;
            elif (paddleup    == 11) t |= 0x10;
        } elif (ay[dest] >= 192)
        {   if   (paddledown  ==  3) t |= 0x80;
            elif (paddledown  ==  6) t |= 0x40;
            elif (paddledown  ==  9) t |= 0x20;
            elif (paddledown  == 11) t |= 0x10;
        }
        if (ax[dest] <=  64)
        {   if   (paddleleft  ==  3) t |= 0x80;
            elif (paddleleft  ==  6) t |= 0x40;
            elif (paddleleft  ==  9) t |= 0x20;
            elif (paddleleft  == 11) t |= 0x10;
        } elif (ax[dest] >= 192)
        {   if   (paddleright ==  3) t |= 0x80;
            elif (paddleright ==  6) t |= 0x40;
            elif (paddleright ==  9) t |= 0x20;
            elif (paddleright == 11) t |= 0x10;
        }
        pviwrite(IE_P1RIGHTKEYS   + (dest * 4), t, FALSE);

        engine_dopaddle(source, dest);

        if (psu & PSU_F) // paddle interpolation bit
        {   pviwrite(PVI_P1PADDLE + dest, (UBYTE) ay[dest], TRUE);
            sy[dest] = (UBYTE) ay[dest];
        } else
        {   pviwrite(PVI_P1PADDLE + dest, (UBYTE) ax[dest], TRUE);
            sx[dest] = (UBYTE) ax[dest];
}   }   }

MODULE __inline void ie_emuinput(void)
{   UBYTE therecv[3],
          thesend[3];

    readkybd();
    ReadJoystick(0);
    ReadJoystick(1);
    // must always do ie_playerinput(foo, 0) then ie_playerinput(foo, 1).
    if (swapped)
    {   ie_playerinput(1, 0);
        ie_playerinput(0, 1);
    } else
    {   ie_playerinput(0, 0);
        ie_playerinput(1, 1);
    }
    p1rumble =
    p2rumble = 0;

    if (recmode != RECMODE_PLAY)
    {   if (KeyDown(console[0]) || (jf[0] & JOYSTART) || (jf[1] & JOYSTART) || console_start) // START /START
        {   t = 0x4F;
        } else
        {   t = 0x0F;
        }
        if (KeyDown(console[1]) || (jf[0] & JOYA    ) || (jf[1] & JOYA    ) || console_a    ) // SELECT/UC
        {   t |= 0x80;
        }
        if (machine == ELEKTOR)
        {   if (KeyDown(console[2]) || (jf[0] & JOYB) || (jf[1] & JOYB    ) || console_b    ) // -/LC
            {   t |= 0x20;
            }
            if (KeyDown(console[3]) || console_reset) // -/RESET (soft)
            {   t |= 0x10;
        }   }

        pviwrite(IE_CONSOLE, t, FALSE);

        domouse();
    }

    if (connected == NET_SERVER)
    {   // we are left player

        thesend[0] = ((memory[IE_P1LEFTKEYS  ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                   |  (memory[IE_P1MIDDLEKEYS] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        thesend[1] =  (memory[IE_P1RIGHTKEYS ] & 0x0F)        // bits 3..0 -> bits  3..0 (masking not really needed)
                   |  (memory[IE_CONSOLE     ] & 0xF0);       // bits 7..4 -> bits  7..4 (masking not really needed)
        thesend[2] =   pviread(PVI_P1PADDLE);                 // bits 7..0 -> bits  7..0

        if (NetExchange((char *) therecv, (const char *) thesend, 3))
        {   pviwrite(IE_P2LEFTKEYS  , (UBYTE) (((therecv[0] & 0xF0) >> 4) | 0x0F), FALSE); // bits 7..4 -> bits 3..0
            pviwrite(IE_P2MIDDLEKEYS, (UBYTE) (( therecv[0] & 0x0F      ) | 0x0F), FALSE); // bits 3..0 -> bits 3..0
            pviwrite(IE_P2RIGHTKEYS , (UBYTE) (( therecv[1] & 0x0F      ) | 0x0F), FALSE); // bits 3..0 -> bits 3..0 (masking not really needed)
            pviwrite(pviread(PVI_P2PADDLE), therecv[2], TRUE                            ); // bits 7..0 -> bits 7..0
            pviwrite(IE_CONSOLE     , (UBYTE) (( therecv[1] & 0xF0      ) | memory[IE_CONSOLE]), FALSE);
    }   }
    elif (connected == NET_CLIENT)
    {   // we are right player

        thesend[0] = ((memory[IE_P2LEFTKEYS  ] & 0x0F) <<  4) // bits 3..0 -> bits  7..4 (masking not really needed)
                   |  (memory[IE_P2MIDDLEKEYS] & 0x0F);       // bits 3..0 -> bits  3..0 (masking not really needed)
        thesend[1] =  (memory[IE_P2RIGHTKEYS ] & 0x0F)        // bits 3..0 -> bits  3..0 (masking not really needed)
                   |  (memory[IE_CONSOLE     ] & 0xF0);       // bits 7..4 -> bits  7..4 (masking not really needed)
        thesend[2] =   pviread(PVI_P2PADDLE);                // bits 7..0 -> bits  7..0

        if (NetExchange((char *) therecv, (const char *) thesend, 3))
        {   pviwrite(IE_P1LEFTKEYS  , (UBYTE) (((therecv[0] & 0xF0) >> 4) | 0x0F), FALSE); // bits 7..4 -> bits 3..0
            pviwrite(IE_P1MIDDLEKEYS, (UBYTE) (( therecv[0] & 0x0F      ) | 0x0F), FALSE); // bits 3..0 -> bits 3..0
            pviwrite(IE_P1RIGHTKEYS , (UBYTE) (( therecv[1] & 0x0F      ) | 0x0F), FALSE); // bits 3..0 -> bits 3..0 (masking not really needed)
            pviwrite(pviread(PVI_P1PADDLE), therecv[2], TRUE                             ); // bits 7..0 -> bits 7..0
            pviwrite(IE_CONSOLE     , (UBYTE) (( therecv[1] & 0xF0      ) | memory[IE_CONSOLE]), FALSE);
    }   }

    if (recmode == RECMODE_PLAY)
    {   pviwrite(IE_CONSOLE, IOBuffer[offset++], FALSE);
    } elif (recmode == RECMODE_RECORD)
    {   // assert(MacroHandle);

        OutputBuffer[0] =   pviread(PVI_P1PADDLE);
        OutputBuffer[1] =  (memory[IE_P1LEFTKEYS  ] & 0xF0)         // bits 7..4 -> bits  7..4 (masking not really needed)
                        | ((memory[IE_P1MIDDLEKEYS] & 0xF0) >>  4); // bits 7..4 -> bits  3..0 (masking not really needed)
        OutputBuffer[2] =   memory[IE_P1RIGHTKEYS ];                // 4 bits are wasted
        OutputBuffer[3] =   pviread(PVI_P2PADDLE);
        OutputBuffer[4] =  (memory[IE_P2LEFTKEYS  ] & 0xF0)         // bits 7..4 -> bits  7..4 (masking not really needed)
                        | ((memory[IE_P2MIDDLEKEYS] & 0xF0) >>  4); // bits 7..4 -> bits  3..0 (masking not really needed)
        OutputBuffer[5] =   memory[IE_P2RIGHTKEYS ];                // 4 bits are wasted
        OutputBuffer[6] =   memory[IE_CONSOLE     ];                // 6 (Interton) or 4 (Elektor) bits are wasted

        DISCARD fwrite(OutputBuffer, 7, 1, MacroHandle);
        /* fwrite() should really check return code
        7 bytes per frame * 50 frames per second = 350 bytes/sec.
        Plus, for Elektor, 1 byte every time the E_RANDOMn hardware
        registers are read. */
    }

    if (console_start) console_start--;
    if (console_a    ) console_a--;
    if (console_b    ) console_b--;
    if (console_reset) console_reset--;
}

EXPORT void the_drawline(int colour, int y)
{   PERSIST int x; // PERSISTent for speed

    if
    (   (machine == INTERTON || memmap == MEMMAP_F)
     && (memory[IE_NOISE] & 0x20)
    )
    {   if (colour >= 8)
        {   colour = 15 - colour; // 0..7
        } else
        {   colour =  7 - colour; // 0..7
    }   }

#ifdef WIN32
    for (x = 0; x < machines[machine].width; x++)
    {   local_drawpixel(x, y, colour);
    }
#endif
#ifdef AMIGA
    for (x = 0; x < machines[machine].width; x++)
    {   screen[x][y] = (UBYTE) colour;
    }
    y *= size;
    for (x = 0; x < divided * size; x++) // this assumes all rows are contiguous
    {   *(longptr[y] + x) = longpens[colour];
    }
#endif
}

EXPORT void ie_inputmirrors(int address)
{   // write mirrors
    if (    mirror_w[address] >= 0x1E81 && mirror_w[address] <= 0x1EFF)
    {   memflags[address] |= MIRRORED_W;
        mirror_w[address] = 0x1E80;
    }
    if (             address  == 0x1E80) // only applies to real $1E80, not mirrors thereof!
    {   memflags[address] |= NOREAD;
    }

    // read mirrors
    if (    mirror_r[address] >= 0x1E81 && mirror_r[address] <= 0x1E97)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1E98 && mirror_r[address] <= 0x1E9B)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x10;
    } elif (mirror_r[address] >= 0x1E9C && mirror_r[address] <= 0x1EA7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1EA8 && mirror_r[address] <= 0x1EAE)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x20;
    } elif (mirror_r[address] >= 0x1EAF && mirror_r[address] <= 0x1EB7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1EB8 && mirror_r[address] <= 0x1EBB)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x30;
    } elif (mirror_r[address] >= 0x1EBC && mirror_r[address] <= 0x1EC7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1EC8 && mirror_r[address] <= 0x1ECE)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x40;
    } elif (mirror_r[address] >= 0x1ECF && mirror_r[address] <= 0x1ED7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1ED8 && mirror_r[address] <= 0x1EDB)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x50;
    } elif (mirror_r[address] >= 0x1EDC && mirror_r[address] <= 0x1EE7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1EE8 && mirror_r[address] <= 0x1EEE)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x60;
    } elif (mirror_r[address] >= 0x1EEF && mirror_r[address] <= 0x1EF7)
    {   memflags[address] |= NOWRITE;
    } elif (mirror_r[address] >= 0x1EF8 && mirror_r[address] <= 0x1EFB)
    {   memflags[address] |= MIRRORED_R;
        mirror_r[address] -= 0x70;
    } elif (mirror_r[address] >= 0x1EFC && mirror_r[address] <= 0x1EFF)
    {   memflags[address] |= NOWRITE;
}   }

EXPORT void ie_viewscreen(void)
{   // assert(machine == INTERTON || machine == ELEKTOR);

    UWORD where = 0x1F80;
    UBYTE data;
    int   i, j;

    T1; textpen(TEXTPEN_VIEW);

    for (j = 0; j < 20; j++)
    {   zprintf
        (   "$%04X: %02X %02X ",
            where,
            memory[where],
            memory[where + 1]
        );

        data = memory[where++];
        for (i = 0; i < 8; i++)
        {   if (data & (128 >> i))
            {   zprintf("#");
            } else
            {   zprintf(".");
        }   }
        zprintf(" ");
        data = memory[where++];
        for (i = 0; i < 8; i++)
        {   if (data & (128 >> i))
            {   zprintf("#");
            } else
            {   zprintf(".");
        }   }
        zprintf("\n");
    }

    zprintf
    (   "$1FA8: %02X %02X %02X %02X %02X\n\n",
        memory[0x1FA8],
        memory[0x1FA9],
        memory[0x1FAA],
        memory[0x1FAB],
        memory[0x1FAC]
    );
    T2;
}

EXPORT void pvi_memmap(void)
{   int i;

    for (i = 0; i <= 255; i++)
    {   memflags[pvibase + i] = 0;
        mirror_r[pvibase + i] =
        mirror_w[pvibase + i] = (UWORD) (pvibase + i);

        if
        (
         // legal: $00..        $2D     hardware registers, RAM, h/w regs, RAM, h/w regs
            (i >= 0x2E && i <= 0x3F) // unmapped
         // legal: $40..        $6D     hardware registers, RAM
         || (i >= 0x6E && i <= 0x7F) // unmapped
         // legal: $80..        $AD     grid, RAM
         || (i >= 0xAE && i <= 0xBF) // unmapped
         // legal: $C0..        $C3     hardware registers
         || (i >= 0xC4 && i <= 0xC5) // unused
         // legal: $C6..        $C9     hardware registers
         ||  i >= 0xCA               // R/- hardware registers, unused, mirrors (not occurring), semi-mirrors
        )
        {   memflags[pvibase + i] |= NOWRITE;
        }

        if
        (
         // legal: $00..        $2D     hardware registers, RAM, h/w regs, RAM, h/w regs
            (i >= 0x2E && i <= 0x3F) // unmapped
         // legal: $40..        $6D     hardware registers, RAM
         || (i >= 0x6E && i <= 0x7F) // unmapped
         // legal: $80..        $AD     grid, RAM
         || (i >= 0xAE && i <= 0xC9) // unmapped, -/W (write-only) and -/- (unusable) hardware registers
         // legal: $CA..        $CD     R/- (read-only) hardware registers
         || (i >= 0xCE && i <= 0xD9) // unused, mirrors (not occurring)
         // legal: $DA..        $DB     semi-mirrors
         || (i >= 0xDC && i <= 0xE9) // mirrors (not occurring)
         // legal: $EA..        $EB     semi-mirrors
         || (i >= 0xEC && i <= 0xF9) // mirrors (not occurring)
         // legal: $FA..        $FB     semi-mirrors
         ||  i >= 0xFC               // mirrors (not occurring)
        )
        {   memflags[pvibase + i] |= NOREAD;
        }

        if
        (   (i >= 0xCA && i <= 0xCB)
         || (i >= 0xDA && i <= 0xDB)
         || (i >= 0xEA && i <= 0xEB)
         || (i >= 0xFA && i <= 0xFB)
        )
        {   memflags[pvibase + i] |= READONCE; // semi-mirrors
        }

        if (i == PVI_PITCH)
        {   memflags[pvibase + i] |= AUDIBLE;
}   }   }

// one (only!) of these must be #defined.
#define MASKVAR // faster for WinArcadia, not much effect for AmiArcadia
// #define MASKMACRO
MODULE void pvi_collisions(int rotations)
{   PERSIST UBYTE t1,
                  t2;
    PERSIST int
#ifdef MASKVAR
                  mask1, // faster than using UWORDs
                  mask2,
                  mask3,
                  mask4,
                  mask5,
                  mask6,
                  mask7,
                  mask8,
                  mask9,
                  mask10,
#endif
                  x; // all PERSISTent for speed

    /* Collisions are not detected between the different sets of sprites,
       only within each set (ie. within each PVI chip). This is presumably
       the correct behaviour.

       ESS-011-7-OmegaLanding and ESS-011-D-Painting seem to expect
       collisions at the far left (< 9) to not register. */

    t1     =
    t2     = 0;
#ifdef MASKVAR
    mask1  =  0x30 << rotations;
    mask2  =  0x50 << rotations;
    mask3  =  0x90 << rotations;
    mask4  =  0x60 << rotations;
    mask5  =  0xA0 << rotations;
    mask6  =  0xC0 << rotations;
    mask7  = (0x10 << rotations) | 8;
    mask8  = (0x20 << rotations) | 8;
    mask9  = (0x40 << rotations) | 8;
    mask10 = (0x80 << rotations) | 8;
#endif
#ifdef MASKMACRO
    #define mask1   (0x30 << rotations)
    #define mask2   (0x50 << rotations)
    #define mask3   (0x90 << rotations)
    #define mask4   (0x60 << rotations)
    #define mask5   (0xA0 << rotations)
    #define mask6   (0xC0 << rotations)
    #define mask7  ((0x10 << rotations) | 8)
    #define mask8  ((0x20 << rotations) | 8)
    #define mask9  ((0x40 << rotations) | 8)
    #define mask10 ((0x80 << rotations) | 8)
#endif

    for (x = 9; x < PVI_XSIZE; x++)
    {   if (!colltable[rast][x]) // just a speed optimization
        {   continue;
        }
        if ((colltable[rast][x] & mask1) == mask1) // #0 & #1 (%0011)
        {   t1 |= 32;
            if (p1sprcol[0]) p1rumble = p1sprcol[0];
            if (p2sprcol[0]) p2rumble = p2sprcol[0];
        }
        if ((colltable[rast][x] & mask2) == mask2) // #0 & #2 (%0101)
        {   t1 |= 16;
            if (p1sprcol[1]) p1rumble = p1sprcol[1];
            if (p2sprcol[1]) p2rumble = p2sprcol[1];
        }
        if ((colltable[rast][x] & mask3) == mask3) // #0 & #3 (%1001)
        {   t1 |=  8;
            if (p1sprcol[2]) p1rumble = p1sprcol[2];
            if (p2sprcol[2]) p2rumble = p2sprcol[2];
        }
        if ((colltable[rast][x] & mask4) == mask4) // #1 & #2 (%0110)
        {   t1 |=  4;
            if (p1sprcol[3]) p1rumble = p1sprcol[3];
            if (p2sprcol[3]) p2rumble = p2sprcol[3];
        }
        if ((colltable[rast][x] & mask5) == mask5) // #1 & #3 (%1010)
        {   t1 |=  2;
            if (p1sprcol[4]) p1rumble = p1sprcol[4];
            if (p2sprcol[4]) p2rumble = p2sprcol[4];
        }
        if ((colltable[rast][x] & mask6) == mask6) // #2 & #3 (%1100)
        {   t1 |=  1;
            if (p1sprcol[5]) p1rumble = p1sprcol[5];
            if (p2sprcol[5]) p2rumble = p2sprcol[5];
        }

        // all these parentheses are necessary, because of order of evaluation of bitwise & and |
        // Inverted force feedback is only implemented for Arcadia (and only for sprite-bg collisions)
        if ((colltable[rast][x] & mask7) == mask7) // #0 & bg
        {   t2 |= 128;
            collx = x; colly = rast;
            if (p1bgcol[0] > 0) p1rumble =  p1bgcol[0];
            if (p2bgcol[0] > 0) p2rumble =  p2bgcol[0];
        }
        if ((colltable[rast][x] & mask8) == mask8) // #1 & bg
        {   t2 |= 64;
            collx = x; colly = rast;
            if (p1bgcol[1] > 0) p1rumble =  p1bgcol[1];
            if (p2bgcol[1] > 0) p2rumble =  p2bgcol[1];
        }
        if ((colltable[rast][x] & mask9) == mask9) // #2 & bg
        {   t2 |= 32;
            collx = x; colly = rast;
            if (p1bgcol[2] > 0) p1rumble =  p1bgcol[2];
            if (p2bgcol[2] > 0) p2rumble =  p2bgcol[2];
        }
        if ((colltable[rast][x] & mask10) == mask10) // #3 & bg
        {   t2 |= 16;
            collx = x; colly = rast;
            if (p1bgcol[3] > 0) p1rumble =  p1bgcol[3];
            if (p2bgcol[3] > 0) p2rumble =  p2bgcol[3];
    }   }

    if (t1) // just a speed optimization
    {   pviwrite(PVI_SPRITECOLLIDE , (UBYTE) (pviread(PVI_SPRITECOLLIDE ) | t1), TRUE);
        pviwrite(PVI_SPRITECOLLIDE2, (UBYTE) (pviread(PVI_SPRITECOLLIDE2) | t1), TRUE);
        pviwrite(PVI_SPRITECOLLIDE3, (UBYTE) (pviread(PVI_SPRITECOLLIDE3) | t1), TRUE);
        pviwrite(PVI_SPRITECOLLIDE4, (UBYTE) (pviread(PVI_SPRITECOLLIDE4) | t1), TRUE);
    }
    if (t2) // just a speed optimization
    {   pviwrite(PVI_BGCOLLIDE ,     (UBYTE) (pviread(PVI_BGCOLLIDE     ) | t2), TRUE);
        pviwrite(PVI_BGCOLLIDE2,     (UBYTE) (pviread(PVI_BGCOLLIDE2    ) | t2), TRUE);
        pviwrite(PVI_BGCOLLIDE3,     (UBYTE) (pviread(PVI_BGCOLLIDE3    ) | t2), TRUE);
        pviwrite(PVI_BGCOLLIDE4,     (UBYTE) (pviread(PVI_BGCOLLIDE4    ) | t2), TRUE);
}   }

MODULE __inline void drawsprites(int whichpvi)
{   PERSIST int whichsprite; // PERSISTent for speed

    // assert(rast < 272);

    for (whichsprite = 0; whichsprite < 4; whichsprite++)
    {   if
        (   sprite[whichpvi][whichsprite].majorrow == -1
         && sprite[whichpvi][whichsprite].starty   == rast
         && rast <= latestsprite
        )
        {   newsprite(whichpvi, whichsprite);
        }
        if (sprite[whichpvi][whichsprite].majorrow != -1)
        {   drawminorrow(whichpvi, whichsprite);
            newminorrow(whichpvi, whichsprite);
}   }   }

MODULE void newsprite(int whichpvi, int whichsprite)
{
#ifdef NOHIDE
    sprite[whichpvi][whichsprite].previmagery = 0;
#endif

    /* It is probable that mid-sprite changes to SIZES are ineffective
       (this behaviour helps HUNTING, SHOOTGAL). */

    sprite[whichpvi][whichsprite].majorrow =
    sprite[whichpvi][whichsprite].minorrow = 0;
    switch (whichsprite)
    {
    case 0:
        sprite[whichpvi][0].size = 1 << ( pviread(PVI_SIZES)       & 3);
    acase 1:
        sprite[whichpvi][1].size = 1 << ((pviread(PVI_SIZES) >> 2) & 3);
    acase 2:
        sprite[whichpvi][2].size = 1 << ((pviread(PVI_SIZES) >> 4) & 3);
    acase 3:
        sprite[whichpvi][3].size = 1 << ((pviread(PVI_SIZES) >> 6) & 3);
    }

    newmajorrow(whichpvi, whichsprite);
}

MODULE void newmajorrow(int whichpvi, int whichsprite)
{   /* These are reread at the start of each major row:
       sprite colour
       sprite imagery
       sprite X-coordinate */

    // assert(minorrow == 0);

    switch (whichsprite)
    {
    case 0:
        sprite[whichpvi][0].colour = ((pviread(PVI_SPR01COLOURS) >> 3) & 7);
    acase 1:
        sprite[whichpvi][1].colour = ( pviread(PVI_SPR01COLOURS)       & 7);
    acase 2:
        sprite[whichpvi][2].colour = ((pviread(PVI_SPR23COLOURS) >> 3) & 7);
    acase 3:
        sprite[whichpvi][3].colour = ( pviread(PVI_SPR23COLOURS)       & 7);
    }

    sprite[whichpvi][whichsprite].imagery = pviread(pvi_spritedata[whichsprite]
                                          + sprite[whichpvi][whichsprite].majorrow);

#ifdef NOHIDE
    sprite[whichpvi][whichsprite].previmagery |= sprite[whichpvi][whichsprite].imagery;
    if (sprite[whichpvi][whichsprite].previmagery == 0)
    {   sprite[whichpvi][whichsprite].imagery = 0xFF;
    }
#endif

    switch (machine)
    {
    case MALZAK:
        if (sprite[whichpvi][whichsprite].dup)
        {   sprite[whichpvi][whichsprite].x = pviread(pvi_spritedata[whichsprite] + 11) - 3; // SPRITEnBX
        } else
        {   sprite[whichpvi][whichsprite].x = pviread(pvi_spritedata[whichsprite] + 10) - 3 + offsetsprites; // SPRITEnAX
        }
    acase GALAXIA:
        if (sprite[whichpvi][whichsprite].dup)
        {   sprite[whichpvi][whichsprite].x = pviread(pvi_spritedata[whichsprite] + 11) - 11; // SPRITEnBX
        } else
        {   sprite[whichpvi][whichsprite].x = pviread(pvi_spritedata[whichsprite] + 10) - 11; // SPRITEnAX
        }
    adefault: // eg. INTERTON, ELEKTOR, ASTROWARS
        if (sprite[whichpvi][whichsprite].dup)
        {   sprite[whichpvi][whichsprite].x = pviread(pvi_spritedata[whichsprite] + 11) + 1; // SPRITEnBX
        } else
        {   sprite[whichpvi][whichsprite].x = pviread(pvi_spritedata[whichsprite] + 10) + 1; // SPRITEnAX
}   }   }

MODULE void newminorrow(int whichpvi, int whichsprite)
{   sprite[whichpvi][whichsprite].minorrow++;
    if (sprite[whichpvi][whichsprite].minorrow == sprite[whichpvi][whichsprite].size)
    {   sprite[whichpvi][whichsprite].minorrow = 0;
        sprite[whichpvi][whichsprite].majorrow++;

        if (sprite[whichpvi][whichsprite].majorrow == 10)
        {   sprite[whichpvi][whichsprite].majorrow = -1;
            // old sprite has been drawn, now we reread SPRITEnBY
            // this must happen now (before running the CPU), not later!
            sprite[whichpvi][whichsprite].starty = rast + ((UBYTE) (pviread(pvi_spritedata[whichsprite] + 13) + 1)) + 1; // SPRITEnBY

            pviwrite(PVI_BGCOLLIDE , (UBYTE) (pviread(PVI_BGCOLLIDE ) | (8 >> whichsprite)), TRUE);
            pviwrite(PVI_BGCOLLIDE2, (UBYTE) (pviread(PVI_BGCOLLIDE2) | (8 >> whichsprite)), TRUE);
            pviwrite(PVI_BGCOLLIDE3, (UBYTE) (pviread(PVI_BGCOLLIDE3) | (8 >> whichsprite)), TRUE);
            pviwrite(PVI_BGCOLLIDE4, (UBYTE) (pviread(PVI_BGCOLLIDE4) | (8 >> whichsprite)), TRUE);

            if
            (   (!(psu & PSU_II) || forceinterrupt >= 2) // forceinterrupt of 2
             && machine != GALAXIA
            )
            {   zinterrupt = TRUE;

#ifdef DEBUGINTERRUPTS
                T1; textpen(TEXTPEN_VERBOSE);
                zprintf
                (   "Generated %s sprite #%d display complete interrupt at rastline %d.\n",
                    sprite[whichpvi][whichsprite].dup ? "duplicate" : "original",
                    whichsprite,
                    rast
                ); T2;
            } else
            {   T1; textpen(TEXTPEN_VERBOSE);
                zprintf
                (   "Suppressed %s sprite #%d display complete interrupt at rastline %d.\n",
                    sprite[whichpvi][whichsprite].dup ? "duplicate" : "original",
                    whichsprite,
                    rast
                ); T2;
#endif

            }
            sprite[whichpvi][whichsprite].dup = TRUE;
        } else
        {   newmajorrow(whichpvi, whichsprite);
}   }   }

MODULE void drawminorrow(int whichpvi, int whichsprite)
{   PERSIST int   expand,
                  i,
                  rast2,
                  x;
    PERSIST UBYTE c; // all PERSISTent for speed

    if (machine == ASTROWARS)
    {   xx = sprite[whichpvi][whichsprite].x / 3 * 4; // 10 PVI = 13.3' non-PVI (implemented as 13)
    } else
    {   xx = sprite[whichpvi][whichsprite].x;
    }
#ifdef NOHIDE
    if (xx >= machines[machine].width)
    {   xx -= machines[machine].width;
    }
#endif

    for (i = 0; i < 8; i++) // the 8 columns of the sprite imagery
    {   if (sprite[whichpvi][whichsprite].imagery & (128 >> i))
        {   expand = TRUE;
            for (x = 0; x < sprite[whichpvi][whichsprite].size; x++, xx++)
            {   if (xx >= PVI_XSIZE)
                {   return;
                } // implied else
                switch (whichpvi)
                {
                case 0:
                    colltable[rast][xx] |= (  0x10 << whichsprite) | sprite[whichpvi][whichsprite].colour;
                acase 1:
                    colltable[rast][xx] |= ( 0x100 << whichsprite) | sprite[whichpvi][whichsprite].colour;
                acase 2:
                    colltable[rast][xx] |= (0x1000 << whichsprite) | sprite[whichpvi][whichsprite].colour;
                }
                if (xx < HIDDEN_X)
                {   continue;
                } // implied else
                switch (machine)
                {
                case INTERTON:
                case ELEKTOR:
                    c = colltable[rast][xx] & 7;
                    rast2 = rast;
                acase ASTROWARS:
                    c = (UBYTE) from_g_x_spr[(colltable[rast][xx] & 7)];
                    rast2 = rast - 16;
                acase GALAXIA:
                    c = (UBYTE) from_g_w_spr[(colltable[rast][xx] & 7)];
                    rast2 = rast - 16;
                acase MALZAK:
                    c = (UBYTE) from_m_spr[(colltable[rast][xx] & 7)];
                    rast2 = rast;
                }

                if
                (   rast2 >= 0
                 && rast2 <  machines[machine].height
                 && xx    <  HIDDEN_X + machines[machine].width
                 && c     != screen[xx - HIDDEN_X][rast2]
                 && (machine != MALZAK || rast >= 5) // kludge to avoid sprites in status area
                 && spritesvisible
                )
                {   local_drawpixel
                    (   xx - HIDDEN_X,
                        rast2,
                        c
                    );
        }   }   }
        else
        {   xx += sprite[whichpvi][whichsprite].size;
            expand = FALSE;
        }
        if (machine == ASTROWARS && (i == 2 || i == 5 || i == 8))
        {   if (expand)
            {   switch (whichpvi)
                {
                case 0:
                    colltable[rast][xx] |= (  0x10 << whichsprite) | sprite[whichpvi][whichsprite].colour;
                acase 1:
                    colltable[rast][xx] |= ( 0x100 << whichsprite) | sprite[whichpvi][whichsprite].colour;
                acase 2:
                    colltable[rast][xx] |= (0x1000 << whichsprite) | sprite[whichpvi][whichsprite].colour;
                }
                if
                (   rast2 >= 0
                 && rast2 <  machines[machine].height
                 && xx    >= HIDDEN_X
                 && xx    <  HIDDEN_X + machines[machine].width
                 && c     != screen[xx - HIDDEN_X][rast2]
                 && spritesvisible
                )
                {   local_drawpixel
                    (   xx - HIDDEN_X,
                        rast2,
                        c
                    );
            }   }
            xx++;
}   }   }

MODULE void doshadows(void)
{   PERSIST int x;

    if (shadowing)
    {   for (x = 0; x < PVI_XSIZE; x++)
        {   if (shady[x] != 16)
            {   local_drawpixel(x, rast, shady[x]);
                shady[x] = 16;
}   }   }   }

MODULE __inline void local_drawpixel(int x, int y, int colour)
{   if (screen[x][y] != colour)
    {   screen[x][y] = (UBYTE) colour;
        drawpixel(x, y, colour);
}   }
